# ----------------------------------------------------------------------------
# PyGMTSAR
# 
# This file is part of the PyGMTSAR project: https://github.com/mobigroup/gmtsar
# 
# Copyright (c) 2022, Alexey Pechnikov
# 
# Licensed under the BSD 3-Clause License (see LICENSE for details)
# ----------------------------------------------------------------------------

class datagrid:
    """
    A class representing a data grid.

    Attributes
    ----------
    chunksize : int
        The chunk size for data compression. Default is 512.
    netcdf_engine : str
        The engine used for NetCDF file operations. Default is 'h5netcdf'.
    netcdf_complevel : int
        The compression level for data compression. Default is 3.
    noindex : np.uint32
        The NODATA index value for transform matrices.
    
    Notes
    ----------
    That's possible to define a special NetCDF backend for Docker environment or other cases:

    if os.path.exists('/.dockerenv') and not 'google.colab' in sys.modules:
        # use different NetCDF backend in Docker containers
        from pygmtsar import datagrid
        datagrid.netcdf_engine = 'netcdf4'
    
    """
    import numpy as np

    # Minimum valid Sentinel-1 radar amplitude from GMTSAR code
    #amplitude_threshold = 5.e-21
    # NetCDF options, see https://docs.xarray.dev/en/stable/user-guide/io.html#zarr-compressors-and-filters
    chunksize = 2048
    netcdf_engine = 'h5netcdf'
    #netcdf_engine = 'netcdf4'
    netcdf_chunksize = 512
    netcdf_compression_algorithm = 'zlib'
    netcdf_complevel = -1
    netcdf_shuffle = True
    netcdf_queue = 16

    # define lost class variables due to joblib via arguments
    def _compression(self, shape=None, chunksize=None):
        """
        Return the compression options for a data grid.

        Parameters
        ----------
        shape : tuple, list, np.ndarray, optional
            The shape of the data grid. Required if chunksize is less than grid dimension sizes. Default is None.
        chunksize : int or tuple, optional
            The chunk size for data compression. If not specified, the class attribute chunksize is used.

        Returns
        -------
        dict
            A dictionary containing the compression options for the data grid.

        Examples
        --------
        Get the compression options for a data grid with shape (1000, 1000):

        >>> compression(shape=(1000, 1000))
        {'zlib': True, 'complevel': 3, 'chunksizes': (512, 512)}

        Get the compression options for a data grid with chunksize 256:

        >>> compression(chunksize=256)
        {'zlib': True, 'complevel': 3, 'chunksizes': (256, 256)}
        """
        import numpy as np

        if chunksize is None:
            chunksize = self.netcdf_chunksize

        assert chunksize is not None, 'compression() chunksize is None'
        if isinstance(chunksize, (tuple, list, np.ndarray)):
            # use as is, it can be 2D or 3D grid (even 1D while it is not used for now)
            if shape is not None:
                assert len(shape) == len(chunksize), f'ERROR: defined shape and chunksize dimensions are not equal: {len(shape)} != {len(chunksize)}'
                chunksizes = tuple([chunksize[dim] if chunksize[dim]<shape[dim] else shape[dim] for dim in range(len(shape))])
            else:
                chunksizes = chunksize
        else:
            if shape is not None:
                # 2D or 3D grid
                chunksizes = []
                for idim in range(len(shape)):
                    chunksizes.append(chunksize if chunksize<shape[idim] else shape[idim])
                # set first dimension chunksize to 1 for 3D array
                if len(chunksizes) == 3:
                    chunksizes[0] = 1
                chunksizes = tuple(chunksizes)
            else:
                chunksizes=(chunksize, chunksize)
        opts = dict(chunksizes=chunksizes)
        if self.netcdf_compression_algorithm is not None and self.netcdf_complevel >= 0:
            opts[self.netcdf_compression_algorithm] = True
            opts['complevel'] = self.netcdf_complevel
            opts['shuffle'] = self.netcdf_shuffle
        return opts

    @staticmethod
    def is_ra(grid):
        """
        Checks if the given grid is in radar coordinates.

        Parameters
        ----------
        grid : xarray.DataArray
            The grid to check.

        Returns
        -------
        bool
            True if the grid is in radar coordinates, False otherwise.
        """
        dims = grid.dims
        if 'y' in dims and 'x' in dims:
            return True
        return False

    @staticmethod
    def is_geo(grid):
        """
        Checks if the given grid is in geographic coordinates.

        Parameters
        ----------
        grid : xarray.DataArray
            The grid to check.

        Returns
        -------
        bool
            True if the grid is in geographic coordinates, False otherwise.
        """
        dims = grid.dims
        if 'lat' in dims and 'lon' in dims:
            return True
        return False
 
    # da.dropna(dim=dim, how='all') is not fast at all
    @staticmethod
    def cropna(das, index=-1):
        """
        Crop the valid extent of a raster by removing rows and columns containing only NODATA values.

        Parameters
        ----------
        das : xarray.DataArray
            The input 2D or 3D grid to be cropped.

        Returns
        -------
        xarray.DataArray
            The cropped 2D or 3D grid.

        Examples
        --------
        Crop the valid extent of a raster:
        stack.cropna(grid)

        Notes
        -----
        This method crops the input grid by removing rows and columns that contain only NODATA values.
        It operates on 2D or 3D grids, where the NODATA values are represented as NaN values.
        The resulting grid has a reduced size, containing only the valid extent of the input grid.
        If the input grid is 3D, the cropping is performed along the dimensions other than 'pair' or 'date'.
        """
        # crop NaNs
        dims = [dim for dim in das.dims if dim != 'pair' and dim != 'date']
        dim0 = [dim for dim in das.dims if dim in ['pair', 'date']]
        #print ('dims', dims, 'dim0', dim0)
        assert len(dims) == 2, 'ERROR: the input should be 3D array with "pair" or "date" coordinate'
        # slow check using all the grids in the stack
        #da = das.min(dim0)
        # fast check using the only "index" grid in the stack
        da = das.isel({dim0[index]: index}) if dim0 != [] else das
        indexer = {}
        for dim in dims:
            da = da.dropna(dim=dim, how='all')
            dim_min, dim_max = da[dim].min().item(), da[dim].max().item()
            indexer[dim] = slice(dim_min, dim_max)
        #print ('indexer', indexer)
        return das.loc[indexer]
    
    # replacement for GMTSAR gaussians
    # gauss5x5 = np.genfromtxt('/usr/local/GMTSAR/share/gmtsar/filters/gauss5x5',skip_header=True)
    # gaussian_kernel(5,1) ~= gauss5x5
    @staticmethod
    def gaussian_kernel(size=(5,5), std=(1,1)):
        """
        Generate a 2D Gaussian kernel matrix.

        Parameters
        ----------
        size : tuple, optional
            The size of the kernel matrix in (rows, columns). Default is (5, 5).
        std : tuple, optional
            The standard deviation of the Gaussian distribution in (row_std, column_std). Default is (1, 1).

        Returns
        -------
        numpy.ndarray
            The 2D Gaussian kernel matrix.

        Examples
        --------
        Generate a 5x5 Gaussian kernel with standard deviation of 1 in both dimensions:
        gaussian_kernel(size=(5, 5), std=(1, 1))

        Generate a 3x3 Gaussian kernel with standard deviation of 0.5 in row dimension and 1 in column dimension:
        gaussian_kernel(size=(3, 3), std=(0.5, 1))
        """
        import numpy as np
        from scipy import signal
        matrix1 = signal.gaussian(size[0], std=std[0]).reshape(size[0], 1)
        matrix2 = signal.gaussian(size[1], std=std[1]).reshape(size[1], 1)
        matrix2d = np.outer(matrix1, matrix2)
        return matrix2d

    @staticmethod
    def get_bounds(geometry):
        import geopandas as gpd
        import xarray as xr
    
        if isinstance(geometry, (xr.DataArray, xr.Dataset)):
            lon_start = geometry.lon.min().item()
            lat_start = geometry.lat.min().item()
            lon_end   = geometry.lon.max().item()
            lat_end   = geometry.lat.max().item()
            bounds = lon_start, lat_start, lon_end, lat_end
        elif isinstance(geometry, gpd.GeoDataFrame):
            bounds = geometry.dissolve().envelope.item().bounds
        elif isinstance(geometry, gpd.GeoSeries):
            bounds = geometry.unary_union.envelope.bounds
        elif isinstance(geometry, tuple):
            # geometry is already bounds
            bounds = geometry
        else:
            bounds = geometry.bounds
        #print ('bounds', bounds)
        #lon_start, lat_start, lon_end, lat_end
        return bounds

#    @staticmethod
#    def nanconvolve2d(data, kernel, threshold=1/3.):
#        """
#        Convolution using generic kernel on a 2D array with NaNs
#        """
#        import numpy as np
#        import scipy.signal
#        # np.complex128 includes np.float64 real and imagine part
#        vals = data.astype(np.complex128)
#        #vals[np.isnan(data)] = 0
#        #vals[~np.isnan(data)] += 1j
#        nanmask = np.isnan(data)
#        vals[nanmask] = 0
#        vals[~nanmask] += 1j
#
#        conv = scipy.signal.convolve2d(vals, 
#                                      kernel.astype(np.complex128), 
#                                      mode='same', boundary='symm'
#                                     )
#        # suppress incorrect division warning
#        np.seterr(invalid='ignore')
#        return np.where(conv.imag >= threshold*np.sum(kernel), np.divide(conv.real, conv.imag), np.nan)

#     @staticmethod
#     def nanconvolve2d_gaussian(dataarray, sigmas, truncate):
#         """
#         Apply lazy convolution using Gaussian kernel on a 2D array with NaN values.
# 
#         Parameters
#         ----------
#         dataarray : xarray.DataArray
#             The input 2D array with NaN values.
#         sigmas : tuple
#             The standard deviations of the Gaussian kernel in (row_sigma, column_sigma).
#         truncate : float
#             The truncation factor for the Gaussian kernel.
# 
#         Returns
#         -------
#         xarray.DataArray
#             The convolved data array with NaN values.
# 
#         Examples
#         --------
#         Apply Gaussian convolution with sigmas (2, 2) and truncate factor 3 to a 2D data array:
#         nanconvolve2d_gaussian(dataarray, sigmas=(2, 2), truncate=3)
#         """
#         import xarray as xr
#         import dask
#         from dask_image.ndfilters import gaussian_filter as dask_gaussian_filter
# 
#         # this command works but it is slow
#         #da = dask.array.where(dask.array.isnan(dataarray), 0, 1j + dataarray)
#         # this command is fast and it replaces nan + 1j to to 0.+0.j
#         da = (dataarray + 1j).fillna(0).data
#         conv = dask_gaussian_filter(da, sigmas, mode='reflect', truncate=truncate)
#         da_conv = xr.DataArray(conv.real/conv.imag, coords=dataarray.coords, name=dataarray.name)
#         return da_conv

    def nearest_grid(self, in_grid, search_radius_pixels=None):
        """
        Perform nearest neighbor interpolation on a 2D grid.

        Parameters
        ----------
        in_grid : xarray.DataArray
            The input 2D grid to be interpolated.
        search_radius_pixels : int, optional
            The interpolation distance in pixels. If not provided, the default is set to the chunksize of the Stack object.

        Returns
        -------
        xarray.DataArray
            The interpolated 2D grid.

        Examples
        --------
        Fill gaps in the specified grid using nearest neighbor interpolation:
        stack.nearest_grid(grid)

        Notes
        -----
        This method performs nearest neighbor interpolation on a 2D grid. It replaces the NaN values in the input grid with
        the nearest non-NaN values. The interpolation is performed within a specified search radius in pixels.
        If a search radius is not provided, the default search radius is set to the chunksize of the Stack object.
        """
        from scipy.spatial import cKDTree
        import xarray as xr
        import numpy as np

        assert in_grid.chunks is not None, 'nearest_grid() input grid chunks are not defined'

        if search_radius_pixels is None:
            search_radius_pixels = self.chunksize
        elif search_radius_pixels <= 0:
            print (f'NOTE: interpolation ignored for search_radius_pixels={search_radius_pixels}')
            return in_grid
        else:
            assert search_radius_pixels <= self.chunksize, \
                f'ERROR: apply nearest_grid_pixels() multiple times to fill gaps more than {self.chunksize} pixels chunk size'

        def func(grid, y, x, distance, scaley, scalex):

            grid1d = grid.reshape(-1).copy()
            nanmask0 = np.isnan(grid1d)
            # all the pixels already defined
            if np.all(~nanmask0):
                return grid

            # crop full grid subset to search for missed values neighbors
            ymin = y.min()-scaley*distance-1
            ymax = y.max()+scaley*distance+1
            xmin = x.min()-scalex*distance-1
            xmax = x.max()+scalex*distance+1
            if self.is_ra(in_grid):
                data = in_grid.sel(y=slice(ymin, ymax), x=slice(xmin, xmax))
                ys, xs = data.y, data.x
            else:
                data = in_grid.sel(lat=slice(ymin, ymax), lon=slice(xmin, xmax))
                ys, xs = data.lat, data.lon
            # compute dask arrays to prevent ineffective index lookup
            ys, xs = [vals.values.reshape(-1) for vals in xr.broadcast(ys, xs)]
            data1d = data.values.reshape(-1)
            nanmask = np.isnan(data1d)
            # all the subset pixels are empty, the search is useless
            if np.all(nanmask):
                return grid

            # build index tree for all the valid subset values
            source_yxs = np.stack([ys[~nanmask]/scaley, xs[~nanmask]/scalex], axis=1)
            tree = cKDTree(source_yxs, compact_nodes=False, balanced_tree=False)

            # query the index tree for all missed values neighbors
            target_yxs = np.stack([(y/scaley).reshape(-1)[nanmask0], (x/scalex).reshape(-1)[nanmask0]], axis=1)
            #assert 0, target_yxs
            d, inds = tree.query(target_yxs, k = 1, distance_upper_bound=distance, workers=1)
            # fill missed values using neighbors when these ones are found
            inds = np.where(np.isinf(d), 0, inds)
            grid1d[nanmask0] = np.where(np.isinf(d), np.nan, data1d[~nanmask][inds])
            return grid1d.reshape(grid.shape)

        coords = ['y', 'x'] if self.is_ra(in_grid) else ['lat', 'lon']
        scale = [in_grid[coord].diff(coord).item(0) for coord in coords]
        yy = xr.DataArray(in_grid[coords[0]]).chunk(-1)
        xx = xr.DataArray(in_grid[coords[1]]).chunk(-1)
        ys, xs = xr.broadcast(yy,xx)

        # xarray wrapper
        grid = xr.apply_ufunc(
            func,
            in_grid,
            ys.chunk(in_grid.chunks),
            xs.chunk(in_grid.chunks),
            dask='parallelized',
            vectorize=False,
            output_dtypes=[np.float32],
            dask_gufunc_kwargs={'distance': search_radius_pixels, 'scaley': scale[0], 'scalex': scale[1]},
        )
        assert grid.chunks is not None, 'nearest_grid() output grid chunks are not defined'
        return grid

    def get_spacing(self, grid=1):
        """
        Compute the ground pixel size in meters for the default processing grid or the defined one.

        Parameters
        ----------
        grid : tuple or xarray.DataArray, optional
            A pair of x, y grid decimation coefficients or a 2D or 3D Xarray DataArray representing the decimation grid.
            The default is (1, 4) for the default processing grid.

        Returns
        -------
        tuple
            The ground pixel size(s) in meters.

        Examples
        --------
        Get the default ground pixel size:
        stack.get_spacing()
        >>> (14.0, 15.7)

        Get the ground pixel size for an unwrapped phase grid with a decimation of {'y': 2, 'x': 2}:
        stack.get_spacing(unwrap)
        >>> (27.9, 29.5)

        Notes
        -----
        This method computes the ground pixel size in meters for the default processing grid or a user-defined grid.
        The pixel size is calculated based on the azimuth and range pixel sizes obtained from the processing parameters.
        If a grid is provided, the pixel sizes are adjusted according to the grid decimation coefficients.
        The pixel sizes are rounded to one decimal place.
        """
        # pixel size in meters
        return self.PRM().get_spacing(grid)

    def get_coarsen(self, factor):
        import numpy as np

        # expand simplified definition
        if not isinstance(factor, (list,tuple, np.ndarray)):
            coarsens = (factor, factor)
        else:
            coarsens = factor

        types = [type(coarsen) for coarsen in coarsens]
        assert types[0] == types[1], f'Mixed coarsen datatypes are not allowed: {types}'
        if types[0] == int:
            # defined in pixels, as required
            return coarsens

        # float coarsen should be defined in meters, convert to pixels
        #print ('coarsens', coarsens)
        psizes = self.get_spacing()
        #print ('psizes', psizes)
        # follow same agreement as Stack.decimator() function for rounding
        if coarsens[1] > 4*psizes[1]:
            coarsens = np.round([coarsens[0]/psizes[0], coarsens[1]/psizes[1]/4]).astype(int)
            coarsens[1] = 4 * coarsens[1]
        else:
            coarsens = np.round([coarsens[0]/psizes[0], coarsens[1]/psizes[1]]).astype(int)

        return coarsens

#     #decimator = lambda da: da.coarsen({'y': 2, 'x': 2}, boundary='trim').mean()
#     def decimator(self, resolution=60, grid=(1, 4), func='mean', debug=False):
#         """
#         Return function for pixel decimation to the specified output resolution.
# 
#         Parameters
#         ----------
#         resolution : int, optional
#             DEM grid resolution in meters. The same grid is used for geocoded results output.
#         grid : tuple, optional
#             Grid size for pixel decimation in the format (vertical, horizontal).
#         debug : bool, optional
#             Boolean flag to print debug information.
# 
#         Returns
#         -------
#         callable
#             Post-processing lambda function.
# 
#         Examples
#         --------
#         Decimate computed interferograms to default DEM resolution 60 meters:
#         decimator = stack.decimator()
#         stack.intf(pairs, func=decimator)
#         """
#         import numpy as np
#         import dask
# 
#         # special cases: scale factor should be 4*N or 2*N to prevent rounding issues
#         # grid can be defined as [] or () or xarray dataarray
#         grid_as_coeffs = isinstance(grid,(list, tuple))
#         if   (grid_as_coeffs and grid==(1, 1)) or (not grid_as_coeffs and grid.x.diff('x')[0].item()) == 1:
#             xscale0 = 4
#         elif (grid_as_coeffs and grid==(1, 2)) or (not grid_as_coeffs and grid.x.diff('x')[0].item()) == 2:
#             xscale0 = 2
#         else:
#             xscale0 = 1
#         if debug:
#             print (f'DEBUG: scale to square grid: xscale0={xscale0}')
# 
#         dy, dx = self.get_spacing(grid)
#         yscale, xscale = int(np.round(resolution/dy)), int(np.round(resolution/dx/xscale0))
#         if debug:
#             print (f'DEBUG: ground pixel size in meters: y={dy}, x={dx}')
#         if yscale <= 1 and xscale <= 1 and xscale0==1:
#             # decimation impossible
#             if debug:
#                 print (f'DEBUG: decimator = lambda da: da')
#             return lambda da: da
#         if debug:
#             print (f"DEBUG: decimator = lambda da: da.coarsen({{'y': {yscale}, 'x': {xscale0*xscale}}}, boundary='trim').{func}()")
# 
#         # decimate function
#         def decimator(da):
#             # workaround for Google Colab when we cannot save grids with x,y coordinate names
#             # also supports geographic coordinates
#             yname = [varname for varname in ['y', 'lat', 'a'] if varname in da.dims][0]
#             xname = [varname for varname in ['x', 'lon', 'r'] if varname in da.dims][0]
#             coarsen_args = {yname: yscale, xname: xscale0*xscale}
#             #if debug:
#             #    print (f"Decimate y variable '{yname}' for scale 1/{yscale} and x variable '{xname}' for scale 1/{xscale}")
#             # avoid creating the large chunks
#             with dask.config.set(**{'array.slicing.split_large_chunks': True}):
#                 if func == 'mean':
#                     return da.coarsen(coarsen_args, boundary='trim').mean()
#                 elif func == 'min':
#                     return da.coarsen(coarsen_args, boundary='trim').min()
#                 elif func == 'max':
#                     return da.coarsen(coarsen_args, boundary='trim').max()
#                 elif func == 'count':
#                     return da.coarsen(coarsen_args, boundary='trim').count()
#                 elif func == 'sum':
#                     return da.coarsen(coarsen_args, boundary='trim').sum()
#                 else:
#                     raise ValueError(f"Unsupported function {func}. Should be 'mean','min','max','count', or 'sum'")
# 
#         # return callback function
#         return lambda da: decimator(da)

#     # anti-aliasing filter and downscaling, double filter (potentially faster)
#     def antialiasing_downscale(self, da, wavelength, truncate=3, coarsen=(1,4)):
#         import xarray as xr
#         import numpy as np
#         import dask
#         from dask_image.ndfilters import gaussian_filter as dask_gaussian_filter
#         # coarse grid to square cells
#         if coarsen is not None:
#             conv = dask_gaussian_filter(da.data, coarsen, mode='reflect', truncate=2)
#             da_square = xr.DataArray(conv, coords=da.coords, name=da.name).coarsen({'y': coarsen[0], 'x': coarsen[1]}, boundary='trim').mean()
#         # antialiasing (multi-looking) filter
#         if wavelength is None:
#             return da_square
#         dy, dx = self.get_spacing()
#         print ('DEBUG dy, dx', dy, dx)
#         sigmas = int(np.round(wavelength/dy/coarsen[0])), int(np.round(wavelength/dx/coarsen[1]))
#         print ('DEBUG sigmas', sigmas)
#         conv = dask_gaussian_filter(da_square.data, sigmas, mode='reflect', truncate=2)
#         return xr.DataArray(conv, coords=da_square.coords, name=da.name)
    # anti-aliasing filter and downscaling, single filter (potentially more accurate)
#     # coarsen = None disables downscaling and uses wavelength to filter
#     # coarsen=1 disables downscaling and use coarsen/cutoff filter
#     def antialiasing_downscale(self, da, weight=None, wavelength=None, coarsen=(1,4), debug=False):
#         import xarray as xr
#         import numpy as np
#         import dask
#         from dask_image.ndfilters import gaussian_filter as dask_gaussian_filter
#         # GMTSAR constant 5.3 defines half-gain at filter_wavelength
#         # https://github.com/gmtsar/gmtsar/issues/411
#         cutoff = 5.3
# 
#         # expand simplified definition
#         if coarsen is not None and not isinstance(coarsen, (list,tuple, np.ndarray)):
#             coarsen = (coarsen, coarsen)
# 
#         # allow this case to save the original grid resolution
#         if wavelength is None and coarsen is None:
#             return da
# 
#         # antialiasing (multi-looking) filter
#         if wavelength is None:
#             sigmas = np.round([coarsen[0]/cutoff, coarsen[1]/cutoff], 2)
#         else:
#             dy, dx = self.get_spacing(da)
#             #print ('DEBUG dy, dx', dy, dx)
#             #sigmas = int(np.round(wavelength/dy/coarsen[0])), int(np.round(wavelength/dx))
#             sigmas = np.round([wavelength/cutoff/dy, wavelength/cutoff/dx], 2)
#         if debug:
#             print ('DEBUG: antialiasing_downscale sigmas', sigmas, 'for specified wavelength', wavelength)
# 
#         # weighted and not weighted convolution
#         if weight is None:
#             if debug:
#                 print ('DEBUG: antialiasing_downscale not weighted filtering')
#             conv = dask_gaussian_filter(da.data, sigmas, mode='reflect', truncate=2)
#         else:
#             assert da.shape == weight.shape, f'Different shapes for raster {da.shape} and weight {weight.shape}'
#             assert da.dims == weight.dims, f'Different dimension names for raster {da.dims} and weight {weight.dims}'
#             if debug:
#                 print ('DEBUG: antialiasing_downscale weighted filtering')
#             #conv = dask_gaussian_filter(((1j + da)*weight).data, sigmas, mode='reflect', truncate=2)
#             # replace nan + 1j to to 0.+0.j
#             da  = ((1j + da) * weight).fillna(0)
#             conv = dask_gaussian_filter(da.data, sigmas, mode='reflect', truncate=2)
#             conv = conv.real/conv.imag
# 
#         da_conv = xr.DataArray(conv, coords=da.coords, name=da.name)
#         # calculate the initial dataarray chunk sizes per dimensions to restore them
#         # it works faster when we prevent small chunks usage
#         chunksizes = (np.max(da.chunksizes['y']), np.max(da.chunksizes['x']))
#         if coarsen is not None:
#             # coarse grid to square cells
#             return da_conv.coarsen({'y': coarsen[0], 'x': coarsen[1]}, boundary='trim').mean().chunk(chunksizes)
#         return da_conv.chunk(chunksizes)



# ----------------------------------------------------------------------------
# PyGMTSAR
# 
# This file is part of the PyGMTSAR project: https://github.com/mobigroup/gmtsar
# 
# Copyright (c) 2023, Alexey Pechnikov
# 
# Licensed under the BSD 3-Clause License (see LICENSE for details)
# ----------------------------------------------------------------------------
from .Stack_incidence import Stack_incidence

class Stack_tidal(Stack_incidence):

#     # naive code is not suitable for large grids and stacks
#     def cube_tidal_los(self, data):
#         """
#         Calculate tidal LOS displacement [m] for data dates and spatial extent
#         """
#         import xarray as xr
# 
#         solid_tide = self.get_tidal()
#         # interpolate on the data_pairs 2D grid
#         tidal_dates = solid_tide.interp_like(data, method='linear', assume_sorted=True)
#         # compute LOS projection, [m]
#         tidal_dates_los = self.los_projection(tidal_dates)
#         return tidal_dates_los

#     def tidal_interp_los(self, dates, grid):
#         """
#         Interpolate pre-calculated tidal displacement for data pairs dates on the specified grid
#         and convert to LOS displacement in meters
#         """
#         import pandas as pd
#         import dask
#         import xarray as xr
#         import numpy as np
# 
#         solid_tide = self.get_tidal().sel(date=dates)
# 
#         # satellite look vector
#         sat_look = self.get_satellite_look_vector()
# 
#         @dask.delayed
#         def interp_block(date, ys_block, xs_block):
#             # use outer variables
#             # interpolate on the data_pairs 2D grid
#             coords = {'y': ys_block, 'x': xs_block}
#             block_tidal = solid_tide.sel(date=date).interp(coords, method='linear', assume_sorted=True)
#             block_look = sat_look.interp(coords, method='linear', assume_sorted=True)
#             los = xr.dot(xr.concat([block_look.look_E, block_look.look_N, block_look.look_U], dim='dim'),
#                       xr.concat([block_tidal.dx, block_tidal.dy, block_tidal.dz], dim='dim'),
#                       dims=['dim'])
#             del block_tidal, block_look, coords
#             return los.data[None,]
# 
#         # define output radar coordinates grid and split to equal chunks and rest
#         ys_blocks = np.array_split(grid.y, np.arange(0, grid.y.size, self.chunksize)[1:])
#         xs_blocks = np.array_split(grid.x, np.arange(0, grid.x.size, self.chunksize)[1:])
# 
#         # per-block processing
#         blocks3d  = []
#         for date in dates:
#             blocks2d  = []
#             for ys_block in ys_blocks:
#                 blocks = []
#                 for xs_block in xs_blocks:
#                     block = dask.array.from_delayed(interp_block(date, ys_block, xs_block),
#                                                     shape=(1, ys_block.size, xs_block.size), dtype=np.float32)
#                     blocks.append(block)
#                     del block
#                 blocks2d.append(blocks)
#                 del blocks
#             blocks3d.append(blocks2d)
#             del blocks2d
#         dask_block = dask.array.block(blocks3d)
#         del blocks3d
# 
#         # convert coordinate to valid dates
#         coords = {'date': pd.to_datetime(dates), 'y': grid.coords['y'], 'x': grid.coords['x']}
#         out = xr.DataArray(dask_block, coords=coords)
#         del dask_block
#         return out.rename('tidal')

    def tidal_los_rad(self, stack):
        """
        Calculate tidal LOS displacement [rad] for data dates and spatial extent
        """
        return 1000*self.tidal_los(stack)/self.los_displacement_mm(1)

#     def tidal_los(self, stack):
#         """
#         Calculate tidal LOS displacement [m] for data_pairs pairs and spatial extent
#         """
#         import xarray as xr
# 
#         # extract pairs
#         pairs, dates = self.get_pairs(stack, dates=True)
#         grid = stack[0] if len(stack.dims) == 3 else stack
#     
#         # interpolate on the data_pairs 2D grid
#         # compute LOS projection, [m]
#         tidal_dates_los = self.tidal_interp_los(dates, grid)
#         # calculate differences between end and start dates for all the pairs
#         tidal_pairs = []
#         for rec in pairs.itertuples():
#             tidal_pair = tidal_dates_los.sel(date=rec.rep) - tidal_dates_los.sel(date=rec.ref)
#             tidal_pairs.append(tidal_pair)
#         # form 3D stack
#         return xr.concat(tidal_pairs, dim='pair').assign_coords({'pair': stack.pair})

#     # faster, but there is no significant advantage vs naive version
#     def stack_tidal_los(self, data_pairs, chunksize=None):
#         """
#         Calculate tidal LOS displacement [m] for data pairs dates and spatial extent
#         """
#         import pandas as pd
#         import dask
#         import xarray as xr
#         import numpy as np
# 
#         if chunksize is None:
#             chunksize = self.chunksize
# 
#         # select only required dates
#         pairs, dates = self.get_pairs(data_pairs, dates=True)
#         data_grid = data_pairs[0] if len(data_pairs.dims) == 3 else data_pairs
# 
#         solid_tide = self.get_tidal().sel(date=dates)
# 
#         # satellite look vector
#         sat_look = self.get_sat_look()
# 
#         @dask.delayed
#         def interp_block(date_ref, date_rep, ys_block, xs_block):
#             # use outer variables
#             # interpolate on the data_pairs 2D grid
#             coords = {'y': ys_block, 'x': xs_block}
#             block_tidal = solid_tide.sel(date=date_rep).interp(coords, method='linear', assume_sorted=True) - \
#                           solid_tide.sel(date=date_ref).interp(coords, method='linear', assume_sorted=True)
#             block_look = sat_look.interp(coords, method='linear', assume_sorted=True)
#             los = xr.dot(xr.concat([block_look.look_E, block_look.look_N, block_look.look_U], dim='dim'),
#                       xr.concat([block_tidal.dx, block_tidal.dy, block_tidal.dz], dim='dim'),
#                       dims=['dim'])
#             del block_tidal, block_look, coords
#             return los.data[None,]
# 
#         # define output radar coordinates grid and split to equal chunks and rest
#         ys_blocks = np.array_split(data_pairs.y, np.arange(0, data_pairs.y.size, chunksize)[1:])
#         xs_blocks = np.array_split(data_pairs.x, np.arange(0, data_pairs.x.size, chunksize)[1:])
# 
#         # per-block processing
#         blocks3d  = []
#         for rec in pairs.itertuples():
#             blocks2d  = []
#             for ys_block in ys_blocks:
#                 blocks = []
#                 for xs_block in xs_blocks:
#                     block = dask.array.from_delayed(interp_block(rec.ref, rec.rep, ys_block, xs_block),
#                                                     shape=(1, ys_block.size, xs_block.size), dtype=np.float32)
#                     blocks.append(block)
#                     del block
#                 blocks2d.append(blocks)
#                 del blocks
#             blocks3d.append(blocks2d)
#             del blocks2d
#         dask_block = dask.array.block(blocks3d)
#         del blocks3d
# 
#         out = xr.DataArray(dask_block, coords=data_pairs.coords)
#         del dask_block
#         return out.rename('tidal')

    def tidal_los(self, stack):
        """
        Interpolate pre-calculated tidal displacement for data pairs dates on the specified grid
        and convert to LOS displacement in meters
        """
        import pandas as pd
        import dask
        import xarray as xr
        import numpy as np

        # extract pairs
        if len(stack.dims) == 3:
            pairs, dates = self.get_pairs(stack, dates=True)
            pairs = pairs[['ref', 'rep']].astype(str).values
            grid = stack[0]
        else:
            dates = [stack[key].dt.date.astype(str).item() for key in ['ref', 'rep']]
            pairs = [dates]
            grid = stack
        #return (pairs, dates)

        solid_tide = self.get_tidal().sel(date=dates)
        # satellite look vector
        sat_look = self.get_satellite_look_vector()

        def interp_block(pair, ys_block, xs_block):
            # use outer variables
            date1, date2 = pair
            # interpolate on the data_pairs 2D grid
            coords = {'y': ys_block, 'x': xs_block}
            block_tidal1 = solid_tide.sel(date=date1).interp(coords, method='linear', assume_sorted=True)\
                .compute(n_workers=1)
            block_tidal2 = solid_tide.sel(date=date2).interp(coords, method='linear', assume_sorted=True)\
                .compute(n_workers=1)
            block_look = sat_look.interp(coords, method='linear', assume_sorted=True)\
                .compute(n_workers=1)
            block_tidal = block_tidal2 - block_tidal1
            los = xr.dot(xr.concat([block_look.look_E, block_look.look_N, block_look.look_U], dim='dim'),
                      xr.concat([block_tidal.dx, block_tidal.dy, block_tidal.dz], dim='dim'),
                      dims=['dim'])
            del block_tidal, block_tidal2, block_tidal1, block_look, coords
            return los.data[None,].astype(np.float32)

        # define output radar coordinates grid and split to equal chunks and rest
        ys_blocks = np.array_split(grid.y, np.arange(0, grid.y.size, self.chunksize)[1:])
        xs_blocks = np.array_split(grid.x, np.arange(0, grid.x.size, self.chunksize)[1:])

        # per-block processing
        blocks3d  = []
        for pair in pairs:
            #print ('pair', pair)
            blocks2d  = []
            for ys_block in ys_blocks:
                blocks = []
                for xs_block in xs_blocks:
                    block = dask.array.from_delayed(dask.delayed(interp_block)(pair, ys_block, xs_block),
                                                    shape=(1, ys_block.size, xs_block.size), dtype=np.float32)
                    blocks.append(block)
                    del block
                blocks2d.append(blocks)
                del blocks
            blocks3d.append(blocks2d)
            del blocks2d
        dask_block = dask.array.block(blocks3d)
        del blocks3d

        if len(stack.dims) == 3:
            out = xr.DataArray(dask_block, coords=stack.coords)
        else:
            out = xr.DataArray(dask_block[0], coords=stack.coords)
        del dask_block
        return out.rename(stack.name)

    def tidal_los_rad(self, stack):
        """
        Calculate tidal LOS displacement [rad] for data_pairs pairs and spatial extent
        """
        return 1000*self.tidal_los(stack)/self.los_displacement_mm(1)

    def tidal_correction_wrap(self, stack):
        """
        Apply tidal correction to wrapped phase pairs [rad] and wrap the result.
        """
        return self.wrap(stack - self.tidal_los_rad(stack)).rename(stack.name)
    
    def get_tidal(self):
        return self.open_cube('tidal')

    def _tidal(self, date, grid):
        import xarray as xr
        import pandas as pd
        import numpy as np
        from io import StringIO, BytesIO
        import subprocess

        coords = np.column_stack([grid.ll.values.ravel(), grid.lt.values.ravel()])
        buffer = BytesIO()
        np.savetxt(buffer, coords, delimiter=' ', fmt='%.6f')
        stdin_data = buffer.getvalue()
        #print ('stdin_data', stdin_data)

        SC_clock_start, SC_clock_stop = self.PRM(date).get('SC_clock_start', 'SC_clock_stop')
        dt = (SC_clock_start + SC_clock_stop)/2
        argv = ['solid_tide', str(dt)]
        #cwd = os.path.dirname(self.filename) if self.filename is not None else '.'
        cwd = self.basedir
        p = subprocess.Popen(argv, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE, cwd=cwd, bufsize=10*1000*1000)
        stdout_data, stderr_data = p.communicate(input=stdin_data)
        stderr_data = stderr_data.decode('utf8')
        if stderr_data is not None and len(stderr_data):
            #print ('DEBUG: solid_tide', stderr_data)
            assert 0, f'DEBUG: solid_tide: {stderr_data}'
        out = np.fromstring(stdout_data, dtype=np.float32, sep=' ').reshape(grid.y.size, grid.x.size, 5)[None,]
        coords = {'date': pd.to_datetime([date]), 'y': grid.y, 'x': grid.x}
        das = {v: xr.DataArray(out[...,idx], coords=coords) for (idx, v) in enumerate(['lon', 'lat', 'dx', 'dy', 'dz'])}
        ds = xr.Dataset(das)
        return ds

    def compute_tidal(self, dates=None, coarsen=32, n_jobs=-1, interactive=False):
        import xarray as xr
        import numpy as np
        from tqdm.auto import tqdm
        import joblib

        if dates is None:
            dates = self.df.index.unique()

        # expand simplified definition
        if not isinstance(coarsen, (list,tuple, np.ndarray)):
            coarsen = (coarsen, coarsen)

        trans_inv = self.get_trans_inv()
        dy, dx = np.diff(trans_inv.y)[0], np.diff(trans_inv.x)[0]
        #print ('dy, dx', dy, dx)
        #step_y, step_x = int(np.round(coarsen[0]*dy)), int(np.round(coarsen[1]*dx))
        # define target grid spacing
        step_y, step_x = int(coarsen[0]/dy), int(coarsen[1]/dx)
        #print ('step_y, step_x', step_y, step_x)
        # fix zero step when specified coarsen is larger than the transform grid coarsen
        if step_y < 1:
            step_y = 1
        if step_x < 1:
            step_x = 1
        grid = trans_inv.sel(y=trans_inv.y[step_y//2::step_y], x=trans_inv.x[step_x//2::step_x])

        def tidal(date):
            return self._tidal(date, grid)

        with self.tqdm_joblib(tqdm(desc='Tidal Computation', total=len(dates))) as progress_bar:
            outs = joblib.Parallel(n_jobs=n_jobs)(joblib.delayed(tidal)(date) for date in dates)

        ds = xr.concat(outs, dim='date')
        if interactive:
            return ds
        self.save_cube(ds, 'tidal', 'Solid Earth Tides Saving')

#     def solid_tide(self, dates, data, debug=False):
#         """
#         Compute the tidal correction using the GMTSAR binary tool solid_tide for geographic coordinate points.
# 
#         Parameters
#         ----------
#         dates : array_like
#             Dates for which to compute the tidal correction. Should be in the format yyyyddd.fffffff.
#         coords : array_like
#             Coordinates for which to compute the tidal correction. Can be a single pair of coordinates [lon, lat],
#             or a list of pairs [[lon1, lat1], [lon2, lat2], ...].
#         debug : bool, optional
#             If True, print debug information. Default is False.
# 
#         Returns
#         -------
#         pandas.DataFrame
#             DataFrame with columns 'lon', 'lat', 'dx', 'dy', 'dz', indexed by 'date'.
# 
#         Examples
#         --------
#         Compute the tidal correction for a single pair of coordinates:
# 
#             coords = stack.solid_tide(stack.df.index, coords=[13.40076, 47.40143])
# 
#         Compute the tidal correction for multiple pairs of coordinates:
# 
#             coords = stack.solid_tide(stack.df.index, coords=[[13.40076, 47.40143], [13.40076, 47.40143]])
#     
#         Compute the tidal correction for point geodataframe:
#             coords = stack.solid_tide(stack.df.index, AOI)
# 
#         Compute the tidal correction for a single record point geodataframe:
#             coords = stack.solid_tide(stack.df.index, AOI.head(1))
# 
#         Output:
# 
#             >>> stack.solid_tide(stack.df.index[:3], coords=[lon, lat])        
#             lon	lat	dx	dy	dz
#             date					
#             2022-06-16	13.400758	47.401431	-0.066918	-0.004765	0.016200
#             2022-06-28	13.400758	47.401431	-0.033571	0.012279	-0.099899
#             2022-07-10	13.400758	47.401431	-0.000806	-0.007983	-0.150675
# 
#             >>> stack.solid_tide(stack.df.index[:3], coords=[[lon, lat], [lon+1, lat+1]])
#             lon	lat	dx	dy	dz
#             date					
#             2022-06-16	13.400758	47.401431	-0.066918	-0.004765	0.016200
#             2022-06-16	14.400758	48.401431	-0.066594	-0.004782	0.010346
#             2022-06-28	13.400758	47.401431	-0.033571	0.012279	-0.099899
#             2022-06-28	14.400758	48.401431	-0.033011	0.012323	-0.100986
#             2022-07-10	13.400758	47.401431	-0.000806	-0.007983	-0.150675
#             2022-07-10	14.400758	48.401431	-0.001107	-0.007758	-0.151605
# 
#         Notes
#         -----
#         This function computes the tidal correction in geographic coordinates based on the dates and coordinates provided.
#         The correction is computed by calling the GMTSAR binary tool 'solid_tide' with the date and coordinates as input.
#         """
#         import numpy as np
#         import geopandas as gpd
#         import pandas as pd
#         from io import StringIO, BytesIO
#         #import os
#         import subprocess
# 
#         if isinstance(data, gpd.GeoDataFrame):
#             coords = np.array([(geom.x, geom.y) for geom in data.geometry])
#         else:
#             coords = np.asarray(data)
#             if len(coords.shape) == 1:
#                 coords = [coords]
#         #print ('coords', coords)
#         buffer = BytesIO()
#         np.savetxt(buffer, coords, delimiter=' ', fmt='%.6f')
#         stdin_data = buffer.getvalue()
#         #print ('stdin_data', stdin_data)
# 
#         outs = []
#         for date in dates:
#             SC_clock_start, SC_clock_stop = self.PRM(None, date).get('SC_clock_start', 'SC_clock_stop')
#             dt = (SC_clock_start + SC_clock_stop)/2
#             argv = ['solid_tide', str(dt)]
#             #cwd = os.path.dirname(self.filename) if self.filename is not None else '.'
#             cwd = self.basedir
#             p = subprocess.Popen(argv, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
#                                  stderr=subprocess.PIPE, cwd=cwd, bufsize=10*1000*1000)
#             stdout_data, stderr_data = p.communicate(input=stdin_data)
# 
#             stderr_data = stderr_data.decode('utf8')
#             if stderr_data is not None and len(stderr_data) and debug:
#                 print ('DEBUG: solid_tide', stderr_data)
#                 return None
# 
#             out = np.fromstring(stdout_data, dtype=float, sep=' ')
#             outs.append(out)
# 
#         return pd.DataFrame(np.asarray(outs).reshape(-1,5),
#                             columns=['lon', 'lat', 'dx', 'dy', 'dz'],
#                             index=np.repeat(dates, len(coords)))
# 
#     # use this direct calculation function to check the new version based on trans_inv (incorrect for now)
#     def solid_tide2(self, dates, data, debug=False):
#         """
#         Compute the tidal correction using the GMTSAR binary tool solid_tide for geographic coordinate points.
# 
#         Parameters
#         ----------
#         dates : array_like
#             Dates for which to compute the tidal correction. Should be in the format yyyyddd.fffffff.
#         coords : array_like
#             Coordinates for which to compute the tidal correction. Can be a single pair of coordinates [lon, lat],
#             or a list of pairs [[lon1, lat1], [lon2, lat2], ...].
#         debug : bool, optional
#             If True, print debug information. Default is False.
# 
#         Returns
#         -------
#         pandas.DataFrame
#             DataFrame with columns 'lon', 'lat', 'dx', 'dy', 'dz', indexed by 'date'.
# 
#         Examples
#         --------
#         Compute the tidal correction for a single pair of coordinates:
# 
#             coords = stack.solid_tide(stack.df.index, coords=[13.40076, 47.40143])
# 
#         Compute the tidal correction for multiple pairs of coordinates:
# 
#             coords = stack.solid_tide(stack.df.index, coords=[[13.40076, 47.40143], [13.40076, 47.40143]])
# 
#         Compute the tidal correction for point geodataframe:
#             coords = stack.solid_tide(stack.df.index, AOI)
# 
#         Compute the tidal correction for a single record point geodataframe:
#             coords = stack.solid_tide(stack.df.index, AOI.head(1))
# 
#         Output:
# 
#             >>> stack.solid_tide(stack.df.index[:3], coords=[lon, lat])        
#             lon	lat	dx	dy	dz
#             date					
#             2022-06-16	13.400758	47.401431	-0.066918	-0.004765	0.016200
#             2022-06-28	13.400758	47.401431	-0.033571	0.012279	-0.099899
#             2022-07-10	13.400758	47.401431	-0.000806	-0.007983	-0.150675
# 
#             >>> stack.solid_tide(stack.df.index[:3], coords=[[lon, lat], [lon+1, lat+1]])
#             lon	lat	dx	dy	dz
#             date					
#             2022-06-16	13.400758	47.401431	-0.066918	-0.004765	0.016200
#             2022-06-16	14.400758	48.401431	-0.066594	-0.004782	0.010346
#             2022-06-28	13.400758	47.401431	-0.033571	0.012279	-0.099899
#             2022-06-28	14.400758	48.401431	-0.033011	0.012323	-0.100986
#             2022-07-10	13.400758	47.401431	-0.000806	-0.007983	-0.150675
#             2022-07-10	14.400758	48.401431	-0.001107	-0.007758	-0.151605
# 
#         Notes
#         -----
#         This function computes the tidal correction in geographic coordinates based on the dates and coordinates provided.
#         The correction is computed by calling the GMTSAR binary tool 'solid_tide' with the date and coordinates as input.
#         """
#         import numpy as np
#         import geopandas as gpd
#         import pandas as pd
#         from io import StringIO, BytesIO
#         #import os
#         import subprocess
# 
#         if isinstance(data, gpd.GeoDataFrame):
#             coords = np.array([(geom.x, geom.y) for geom in data.geometry])
#         else:
#             coords = np.asarray(data)
#             if len(coords.shape) == 1:
#                 coords = [coords]
#         #print ('coords', coords)
#         buffer = BytesIO()
#         np.savetxt(buffer, coords, delimiter=' ', fmt='%.6f')
#         stdin_data = buffer.getvalue()
#         #print ('stdin_data', stdin_data)
# 
#         outs = []
#         for date in dates:
#             SC_clock_start, SC_clock_stop = self.PRM(None, date).get('SC_clock_start', 'SC_clock_stop')
#             dt = (SC_clock_start + SC_clock_stop)/2
#             argv = ['solid_tide', str(dt)]
#             #cwd = os.path.dirname(self.filename) if self.filename is not None else '.'
#             cwd = self.basedir
#             p = subprocess.Popen(argv, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
#                                  stderr=subprocess.PIPE, cwd=cwd, bufsize=10*1000*1000)
#             stdout_data, stderr_data = p.communicate(input=stdin_data)
# 
#             stderr_data = stderr_data.decode('utf8')
#             if stderr_data is not None and len(stderr_data) and debug:
#                 print ('DEBUG: solid_tide', stderr_data)
#                 return None
# 
#             out = np.fromstring(stdout_data, dtype=float, sep=' ')
#             outs.append(out)
# 
#         return pd.DataFrame(np.asarray(outs).reshape(-1,5),
#                             columns=['lon', 'lat', 'dx', 'dy', 'dz'],
#                             index=np.repeat(dates, len(coords)))



# ----------------------------------------------------------------------------
# PyGMTSAR
# 
# This file is part of the PyGMTSAR project: https://github.com/mobigroup/gmtsar
# 
# Copyright (c) 2021, Alexey Pechnikov
# 
# Licensed under the BSD 3-Clause License (see LICENSE for details)
# ----------------------------------------------------------------------------
from .Stack_reframe import Stack_reframe
from .PRM import PRM

class Stack_dem(Stack_reframe):

    buffer_degrees = 0.02

    def get_extent_ra(self):
        """
        minx, miny, maxx, maxy = np.round(geom.bounds).astype(int)
        """
        import numpy as np
        from shapely.geometry import LineString

        dem = self.get_dem()
        df = dem.isel(lon=[0,-1]).to_dataframe().reset_index()
        geom = self.geocode(LineString(np.column_stack([df.lon, df.lat])))
        return geom

#     def get_extent(self, grid=None, subswath=None):
#         import numpy as np
# 
#         bounds = self.get_bounds(self.get_reference(subswath))
#         if grid is None:
#             return bounds
#         return grid\
#                .transpose('lat','lon')\
#                .sel(lat=slice(bounds[1], bounds[3]),
#                     lon=slice(bounds[0], bounds[2]))

    def get_geoid(self, grid=None):
        """
        Get EGM96 geoid heights.

        Parameters
        ----------
        grid : xarray array or dataset, optional
            Interpolate geoid heights on the grid. Default is None.

        Returns
        -------
        None

        Examples
        --------
        stack.get_geoid()

        Notes
        -----
        See EGM96 geoid heights on http://icgem.gfz-potsdam.de/tom_longtime
        """
        import xarray as xr
        import os

        gmtsar_sharedir = PRM().gmtsar_sharedir()
        geoid_filename = os.path.join(gmtsar_sharedir, 'geoid_egm96_icgem.grd')
        geoid = xr.open_dataarray(geoid_filename, engine=self.netcdf_engine, chunks=self.netcdf_chunksize).rename({'y': 'lat', 'x': 'lon'})
        if grid is not None:
            geoid = geoid.interp_like(grid, method='cubic')
        return geoid

    def set_dem(self, dem_filename):
        """
        Set the filename of the digital elevation model (DEM) in WGS84 NetCDF grid.

        Parameters
        ----------
        dem_filename : str or None
            Path to the WGS84 NetCDF DEM file. If provided, the DEM filename will be set. If None, the DEM filename will be cleared.

        Returns
        -------
        self
            Returns the modified instance of the class.

        Examples
        --------
        Set the DEM filename:
        stack = stack.set_dem('data/DEM_WGS84.nc')

        Notes
        -----
        This method sets the filename of the digital elevation model (DEM) data to be used in the SAR processing.
        The DEM file is a NetCDF dataset in geographical coordinates containing elevation values for the Earth's surface.
        By setting the DEM filename, it allows SAR processing algorithms to utilize the DEM data for geocoding, topographic
        correction, and other applications. If `dem_filename` is None, the DEM filename will be cleared.
        """
        import os
        if dem_filename is not None:
            assert os.path.exists(dem_filename), f'DEM file not found: {dem_filename}'
            assert os.path.isfile(dem_filename) and os.access(dem_filename, os.R_OK), f'DEM file is not readable: {dem_filename}'
            self.dem_filename = os.path.relpath(dem_filename,'.')
        else:
            self.dem_filename = None
        return self

    # buffer required to get correct (binary) results from SAT_llt2rat tool
    # small buffer produces incomplete area coverage and restricted NaNs
    # 0.02 degrees works well worldwide but not in Siberia
    # minimum buffer size: 8 arc seconds for 90 m DEM
    # subswath argument is required for aligning
    def get_dem(self, subswath=None):
        """
        Retrieve the digital elevation model (DEM) data.

        Parameters
        ----------
        subswath : str, optional
            Subswath name. Default is None.

        Returns
        -------
        xarray.DataArray
            The DEM data as a DataArray object.

        Raises
        ------
        Exception
            If the DEM is not set before calling this method.

        Examples
        --------
        Get DEM for all the processed subswaths:
        topo_ll = stack.get_dem()

        Get DEM for a single subswath IW1:
        topo_ll = stack.get_dem(1)

        Notes
        -----
        This method retrieves the digital elevation model (DEM) data previously downloaded and stored in a NetCDF file.
        The DEM file is opened, and the elevation variable is extracted. Any missing values in the elevation data are filled
        with zeros (mostly representing water surfaces).
        """
        import xarray as xr
        import os
        import warnings
        # supress warnings "UserWarning: The specified chunks separate the stored chunks along dimension"
        warnings.filterwarnings('ignore')

        if self.dem_filename is None:
            raise Exception('Set DEM first')

        # open DEM file and find the elevation variable
        # because sometimes grid includes 'crs' or other variables
        dem = xr.open_dataset(self.dem_filename, engine=self.netcdf_engine, chunks=self.chunksize)
        if 'lat' not in dem.coords and 'y' in dem.coords:
            dem = dem.rename({'y': 'lat', 'x': 'lon'})
        # define latlon array
        z_array_name = [data_var for data_var in dem.data_vars if len(dem.data_vars[data_var].coords)==2]
        assert len(z_array_name) == 1
        # extract the array and fill missed values (mostly water surfaces)
        dem = dem[z_array_name[0]]
        #.fillna(0)
        # round the coordinates up to 1 mm
        dem['lat'] = dem.lat.round(8)
        dem['lon'] = dem.lon.round(8)

        # crop to reference scene and subswath if defined
        bounds = self.get_bounds(self.get_reference(subswath))
        return dem\
               .transpose('lat','lon')\
               .sel(lat=slice(bounds[1] - self.buffer_degrees, bounds[3] + self.buffer_degrees),
                    lon=slice(bounds[0] - self.buffer_degrees, bounds[2] + self.buffer_degrees))
        #return self.get_extent(dem, subswath)

    def load_dem(self, data, geometry='auto', buffer_degrees=None):
        """
        Load and preprocess digital elevation model (DEM) data from specified datafile or variable.

        Parameters
        ----------
        data : xarray dataarray or str
            DEM filename or variable.

        Returns
        -------
        None

        Examples
        --------
        Load and crop from local NetCDF file:
        stack.load_dem('GEBCO_2020/GEBCO_2020.nc')

        Load and crop from local GeoTIF file:
        stack.load_dem('GEBCO_2019.tif')

        Load from Xarray DataArray or Dataset:
        stack.set_dem(None).load_dem(dem)
        stack.set_dem(None).load_dem(dem.to_dataset())

        Notes
        -----
        This method loads DEM from the user specified file. The data is then preprocessed by removing
        the EGM96 geoid to make the heights relative to the WGS84 ellipsoid.
        """
        import xarray as xr
        import numpy as np
        import rioxarray as rio
        import geopandas as gpd
        import os

        dem_filename = os.path.join(self.basedir, 'DEM_WGS84.nc')

        if self.dem_filename is not None:
            print ('NOTE: DEM exists, ignore the command. Use Stack.set_dem(None) to allow new DEM downloading')
            return

        if buffer_degrees is not None:
            self.buffer_degrees = buffer_degrees

        if isinstance(data, (xr.Dataset)):
            ortho = data[list(data.data_vars)[0]]
        elif isinstance(data, (xr.DataArray)):
            ortho = data
        elif isinstance(data, str) and os.path.splitext(data)[-1] in ['.tiff', '.tif', '.TIF']:
            ortho = rio.open_rasterio(data, chunks=self.chunksize).squeeze(drop=True)\
                .rename({'y': 'lat', 'x': 'lon'})\
                .drop('spatial_ref')
            if ortho.lat.diff('lat')[0].item() < 0:
                ortho = ortho.reindex(lat=ortho.lat[::-1])
        elif isinstance(data, str) and os.path.splitext(data)[-1] in ['.nc', '.netcdf', '.grd']:
            ortho = xr.open_dataarray(data, engine=self.netcdf_engine, chunks=self.chunksize)
        elif isinstance(data, str):
            print ('ERROR: filename extension is not recognized. Should be one from .tiff, .tif, .TIF, .nc, .netcdf, .grd')
        else:
            print ('ERROR: argument is not an Xarray object and it is not a file name')

        # crop to reference scene extent
        geometry_auto = type(geometry) == str and geometry == 'auto'
        bounds = self.get_bounds(self.get_reference() if geometry_auto else geometry)
        ortho = ortho\
               .transpose('lat','lon')\
               .sel(lat=slice(bounds[1] - self.buffer_degrees, bounds[3] + self.buffer_degrees),
                    lon=slice(bounds[0] - self.buffer_degrees, bounds[2] + self.buffer_degrees))

        # heights correction
        geoid = self.get_geoid(ortho)
        if os.path.exists(dem_filename):
            os.remove(dem_filename)
        encoding = {'dem': self._compression(ortho.shape)}
        (ortho + geoid).rename('dem').load()\
            .to_netcdf(dem_filename, encoding=encoding, engine=self.netcdf_engine)

        self.dem_filename = dem_filename

    def download_dem(self, geometry='auto', product='1s'):
        print ('NOTE: Function is removed. Download DEM using Tiles().download_dem()')
        print ('and load with Stack.load_dem() function.')



# ----------------------------------------------------------------------------
# PyGMTSAR
# 
# This file is part of the PyGMTSAR project: https://github.com/mobigroup/gmtsar
# 
# Copyright (c) 2023, Alexey Pechnikov
# 
# Licensed under the BSD 3-Clause License (see LICENSE for details)
# ----------------------------------------------------------------------------
from .Stack_stl import Stack_stl
from .tqdm_dask import tqdm_dask

class Stack_ps(Stack_stl):

    def get_ps(self, name='ps'):
        return self.open_cube(name)

    #from pygmtsar import tqdm_dask
    #Stack.ps = ps    
    #stack.ps(interactive=True)
    #stack.ps()
    #adi = stack.open_grids(None, 'ps')
    #adi
    #ps_decimator = stack.pixel_decimator(resolution=60, grid=adi, debug=True)
    #adi_dec = adi.coarsen({'y': 4, 'x': 16}, boundary='trim').min()
    #adi_dec
    # define PS candidates using Amplitude Dispersion Index (ADI)
    def compute_ps(self, dates=None, data='auto', name='ps', interactive=False):
        import xarray as xr
        import numpy as np
        import dask
        import os
        import warnings
        # suppress Dask warning "RuntimeWarning: invalid value encountered in divide"
        warnings.filterwarnings('ignore')
        warnings.filterwarnings('ignore', module='dask')
        warnings.filterwarnings('ignore', module='dask.core')

        if isinstance(data, str) and data == 'auto':
            # open SLC data as real intensities
            data = np.square(np.abs(self.open_data(dates=dates)))

        # normalize image amplitudes (intensities)
        tqdm_dask(mean := dask.persist(data.mean(dim=['y','x'])), desc='Intensity Normalization')
        # dask.persist returns tuple
        norm = mean[0].mean(dim='date') / mean[0]
        # compute average and std.dev.
        stats = (norm * data).pipe(lambda x: (x.mean(dim='date'), x.std(dim='date')))
        del data, norm
        ds = xr.merge([stats[0].rename('average'), stats[1].rename('deviation'), mean[0].rename('stack_average')])
        del stats, mean
        if interactive:
            return ds
        self.save_cube(ds, name, 'Compute Stability Measures')
        del ds

    def psfunction(self, ps='auto', name='ps'):
        import numpy as np
        if isinstance(ps, str) and ps == 'auto':
            ps = self.get_ps(name)
        psfunction = (ps.average/(2*ps.deviation))
        return psfunction.where(np.isfinite(psfunction)).rename('psf')

    def plot_psfunction(self, psfunction='auto', caption='PS Function', cmap='gray', quantile=None, vmin=None, vmax=None, **kwargs):
        import numpy as np
        import matplotlib.pyplot as plt

        if isinstance(psfunction, str) and psfunction == 'auto':
            psfunction = self.psfunction()

        if quantile is not None:
            assert vmin is None and vmax is None, "ERROR: arguments 'quantile' and 'vmin', 'vmax' cannot be used together"

        if quantile is not None:
            vmin, vmax = np.nanquantile(psfunction, quantile)

        plt.figure()
        psfunction.plot.imshow(cmap=cmap, vmin=vmin, vmax=vmax)
        self.plot_AOI(**kwargs)
        self.plot_POI(**kwargs)
        plt.title(caption)

#     def get_adi_threshold(self, threshold):
#         """
#         Vectorize Amplitude Dispersion Index (ADI) raster values selected using the specified threshold.
#         """
#         import numpy as np
#         import dask
#         import pandas as pd
#         import geopandas as gpd
# 
#         def adi_block(ys, xs):
#             from scipy.interpolate import griddata
#             # we can calculate more accurate later
#             dy = dx = 10
#             trans_inv_block = trans_inv.sel(y=slice(min(ys)-dy,max(ys)+dy), x=slice(min(xs)-dx,max(xs)+dx))
#             lt_block = trans_inv_block.lt.compute(n_workers=1).data.ravel()
#             ll_block = trans_inv_block.ll.compute(n_workers=1).data.ravel()
#             block_y, block_x = np.meshgrid(trans_inv_block.y.data, trans_inv_block.x.data)
#             points = np.column_stack([block_y.ravel(), block_x.ravel()])
#             # following NetCDF indices 0.5,1.5,...
#             adi_block = adi.sel(y=slice(min(ys),max(ys)+1), x=slice(min(xs),max(xs)+1))
#             adi_block_value = adi_block.compute(n_workers=1).data.ravel()
#             adi_block_mask = adi_block_value<=threshold
#             adi_block_value = adi_block_value[adi_block_mask]
#             adi_block_y, adi_block_x = np.meshgrid(adi_block.y, adi_block.x)
#             adi_block_y = adi_block_y.ravel()[adi_block_mask]
#             adi_block_x = adi_block_x.ravel()[adi_block_mask]
#             # interpolate geographic coordinates, coarsen=2 grid is required for the best accuracy
#             grid_lt = griddata(points, lt_block, (adi_block_y, adi_block_x), method='linear').astype(np.float32)
#             grid_ll = griddata(points, ll_block, (adi_block_y, adi_block_x), method='linear').astype(np.float32)
#             # return geographic coordinates and values
#             return np.column_stack([grid_lt, grid_ll, adi_block_value])
#     
#         # data grid and transform table
#         adi = self.get_adi()
#         trans_inv = self.get_trans_inv()
#     
#         # split to equal chunks and rest
#         ys_blocks = np.array_split(np.arange(adi.y.size), np.arange(0, adi.y.size, self.chunksize)[1:])
#         xs_blocks = np.array_split(np.arange(adi.x.size), np.arange(0, adi.x.size, self.chunksize)[1:])
#         # arrays size is unknown so we cannot construct dask array
#         blocks = []
#         for ys_block in ys_blocks:
#             for xs_block in xs_blocks:
#                 block = dask.delayed(adi_block)(ys_block, xs_block)
#                 blocks.append(block)
#                 del block
#     
#         # materialize the result as a set of numpy arrays
#         tqdm_dask(model := dask.persist(blocks), desc='Amplitude Dispersion Index (ADI) Threshold')
#         del blocks
#         # the result is already calculated and compute() returns the result immediately
#         model = np.concatenate(dask.compute(model)[0][0])
#         # convert to geopandas object
#         columns = {'adi': model[:,2], 'geometry': gpd.points_from_xy(model[:,1], model[:,0])}
#         df = gpd.GeoDataFrame(columns, crs="EPSG:4326")
#         del columns
#         return df



# ----------------------------------------------------------------------------
# PyGMTSAR
# 
# This file is part of the PyGMTSAR project: https://github.com/mobigroup/gmtsar
# 
# Copyright (c) 2021, Alexey Pechnikov
# 
# Licensed under the BSD 3-Clause License (see LICENSE for details)
# ----------------------------------------------------------------------------
from .Stack_multilooking import Stack_multilooking

class Stack_landmask(Stack_multilooking):

    def set_landmask(self, landmask_filename):
        """
        Set the landmask file path.

        Parameters
        ----------
        landmask_filename : str or None
            The NetCDF file path of the landmask. If None, the landmask file path will be set to None.

        Examples
        --------
        stack = stack.set_landmask('data/landmask.nc')

        Returns
        -------
        self : Stack
            The Stack object with the updated landmask file path.
        """
        import os
        if landmask_filename is not None:
            assert os.path.exists(landmask_filename), f'Landmask file not found: {landmask_filename}'
            assert os.path.isfile(landmask_filename) and os.access(landmask_filename, os.R_OK), f'Landmask file is not readable: {landmask_filename}'
            self.landmask_filename = os.path.relpath(landmask_filename,'.')
        else:
            self.landmask_filename = None
        return self

    def get_landmask(self):
        """
        Get the landmask grid in geographic coordinates.

        Parameters
        ----------
        None

        Returns
        -------
        landmask : xarray.DataArray
            The landmask grid.

        Examples
        --------
        Get land mask in geographic coordinates:
        landmask = stack.get_landmask()

        Notes
        -----
        This method opens the landmask NetCDF file and extracts the landmask grid.
        """
        import xarray as xr
        import os

        if self.landmask_filename is None:
            raise Exception('Set landmask first')

        # open DEM file and find the elevation variable
        # because sometimes grid includes 'crs' or other variables
        landmask = xr.open_dataset(self.landmask_filename, engine=self.netcdf_engine, chunks=self.chunksize)
        assert 'lat' in landmask.coords and 'lon' in landmask.coords
        # define latlon array
        z_array_name = [data_var for data_var in landmask.data_vars if len(landmask.data_vars[data_var].coords)==2]
        assert len(z_array_name) == 1
        # extract the array and fill missed values by zero (mostly ocean area)
        landmask = landmask[z_array_name[0]].fillna(0)
        # round the coordinates up to 1 mm to have the same grid for dem and landmask
        landmask['lat'] = landmask.lat.round(8)
        landmask['lon'] = landmask.lon.round(8)
        return landmask

    def download_landmask(self, product='1s', debug=False):
        print ('NOTE: Function is removed. Download land mask using Tiles().download_landmask()')
        print ('and load with Stack.load_landmask() function.')

    def load_landmask(self, data, geometry='auto'):
        """
        Load and preprocess land mask data from specified datafile or variable.

        Parameters
        ----------
        data : xarray dataarray or str
            Land mask filename or variable.

        Returns
        -------
        None

        Examples
        --------
        Load and crop from local NetCDF file:
        stack.load_landmask('landmask.nc')

        Load and crop from local GeoTIF file:
        stack.load_landmask('landmask.tif')

        Load from Xarray DataArray or Dataset:
        stack.load_landmask(None).load_landmask(landmask)
        stack.load_landmask(None).load_landmask(landmask.to_dataset())
        
        """
        import xarray as xr
        import numpy as np
        import rioxarray as rio
        import geopandas as gpd
        import os

        # generate the same as DEM grid
        landmask_filename = os.path.join(self.basedir, 'landmask.nc')
        
        if self.landmask_filename is not None:
            print ('NOTE: landmask exists, ignore the command. Use Stack.set_landmask(None) to allow new landmask downloading')
            return

        if isinstance(data, (xr.Dataset)):
            landmask = data[list(data.data_vars)[0]]
        elif isinstance(data, (xr.DataArray)):
            landmask = data
        elif isinstance(data, str) and os.path.splitext(data)[-1] in ['.tiff', '.tif', '.TIF']:
            landmask = rio.open_rasterio(data, chunks=self.chunksize).squeeze(drop=True)\
                .rename({'y': 'lat', 'x': 'lon'})\
                .drop('spatial_ref')
            if landmask.lat.diff('lat')[0].item() < 0:
                landmask = landmask.reindex(lat=landmask.lat[::-1])
        elif isinstance(data, str) and os.path.splitext(data)[-1] in ['.nc', '.netcdf', '.grd']:
            landmask = xr.open_dataarray(data, engine=self.netcdf_engine, chunks=self.chunksize)
        elif isinstance(data, str):
            print ('ERROR: filename extension is not recognized. Should be one from .tiff, .tif, .TIF, .nc, .netcdf, .grd')
        else:
            print ('ERROR: argument is not an Xarray object and it is not a file name')

        # unify to DEM
        dem = self.get_dem()
        landmask = landmask.transpose('lat','lon').reindex_like(dem, method='nearest')

        if os.path.exists(landmask_filename):
            os.remove(landmask_filename)
        encoding = {'landmask': self._compression(landmask.shape)}
        landmask.rename('landmask').load()\
            .to_netcdf(landmask_filename, encoding=encoding, engine=self.netcdf_engine)

        self.landmask_filename = landmask_filename

    def plot_landmask(self, landmask='auto', caption='Land Mask', cmap='binary_r', aspect=None, **kwargs):
        import matplotlib.pyplot as plt

        if isinstance(landmask, str) and landmask == 'auto':
            landmask = self.get_landmask()

        plt.figure()
        landmask.plot.imshow(vmin=0, cmap=cmap)
        self.plot_AOI(**kwargs)
        self.plot_POI(**kwargs)
        if aspect is not None:
            plt.gca().set_aspect(aspect)
        plt.title(caption)



# ----------------------------------------------------------------------------
# PyGMTSAR
#.
# This file is part of the PyGMTSAR project: https://github.com/mobigroup/gmtsar
#
# Copyright (c) 2024, Alexey Pechnikov
#
# Licensed under the BSD 3-Clause License (see LICENSE for details)
# ----------------------------------------------------------------------------
from .Stack_ps import Stack_ps
from .tqdm_dask import tqdm_dask

class Stack_export(Stack_ps):

    def as_geo(self, da):
        """
        Add geospatial attributes (CRS and spatial dimensions) to allow raster operations using RioXarray.

        Parameters
        ----------
        da : xarray.DataArray
            The input 2D or 3D grid to be converted to geospatial.

        Returns
        -------
        xarray.DataArray
            The geospatial 2D or 3D grid.

        Examples
        --------
        Convert a raster to geospatial and mask it using a Shapely vector geometry:
        stack.as_geo(grid).rio.clip([geometry])

        Notes
        -----
        This method adds geospatial attributes (CRS and spatial dimensions) to the input grid,
        allowing raster operations using the RioXarray library. If the input grid is already
        in geographic coordinates, the CRS is set to EPSG 4326 with spatial dimensions 'lat' and 'lon'.
        Otherwise, if the input grid is in radar coordinates, a fake metric coordinate system is used
        with EPSG 3857 and spatial dimensions 'y' and 'x'. The method relies on the availability of the
        'rioxarray' module.
        """
        import rioxarray
        import sys
        #assert 'rioxarray' in sys.modules, 'rioxarray module is not found'
        if self.is_geo(da):
            epsg = 4326
            y_dim = 'lat'
            x_dim = 'lon'
        else:
            # fake metrical coordinate system just to perform spatial operations
            epsg = 3857
            y_dim = 'y'
            x_dim = 'x'
        return da.rio.write_crs(epsg).rio.set_spatial_dims(y_dim=y_dim, x_dim=x_dim)

    @staticmethod
    def as_vtk(dataset):
        """
        Map a 2D image onto a 3D topography.

        Parameters:
            dataset: A file name or Xarray Dataset to use for mapping the image.

        Returns:
            A vtkStructuredGrid representing the image mapped onto the topography.
        """
        from vtk import vtkPoints, vtkStructuredGrid, vtkThreshold, vtkDataObject, \
            VTK_FLOAT, VTK_UNSIGNED_CHAR, vtkStringArray, vtkFloatArray, vtkIntArray
        from vtk.util import numpy_support as vn
        import numpy as np
        import xarray as xr

        assert isinstance(dataset, xr.Dataset), 'ERROR: Expected "dataset" argument as Xarray Dataset'

        # fill NODATA by NAN
        for data_var in dataset.data_vars:
            #if dataset[data_var].values.dtype in [np.dtype('float16'),np.dtype('float32'),np.dtype('float64')]:
            if np.issubdtype(dataset[data_var].dtype, np.floating):
                dataset[data_var].values = dataset[data_var].values.astype('float32')
                if '_FillValue' in dataset[data_var] and not np.isnan(dataset[data_var]._FillValue):
                    dataset[data_var].values[dataset[data_var].values == dataset[data_var]._FillValue] = np.nan

        xs = dataset.x.values
        ys = dataset.y.values
        (yy,xx) = np.meshgrid(ys, xs)
        if 'z' in dataset:
            zs = dataset.z.values
        else:
            # set z coordinate to zero
            zs = np.zeros_like(xx)

        # create raster mask by geometry and for NaNs
        vtk_points = vtkPoints()
        points = np.column_stack((xx.ravel('F'), yy.ravel('F'), zs.ravel('C')))
        vtk_points.SetData(vn.numpy_to_vtk(points, deep=True))

        sgrid = vtkStructuredGrid()
        sgrid.SetDimensions(len(xs), len(ys), 1)
        sgrid.SetPoints(vtk_points)

        for data_var in dataset.data_vars:
            if 'z' == data_var:
                # variable is already included into coordinates
                continue
            da = dataset[data_var]
            dims = da.dims
            #print (data_var, dims, da.dtype)
            if 'band'in dims:
                bands = da.band.shape[0]
                #print ('bands', bands)
                if bands in [3,4]:
                    # RGB or RGBA, select 3 bands only
                    array = vn.numpy_to_vtk(da[:3].round().astype(np.uint8).values.reshape(3,-1).T, deep=True, array_type=VTK_UNSIGNED_CHAR)
                    array.SetName(da.name)
                    sgrid.GetPointData().AddArray(array)
                elif bands == 1:
                    array = vn.numpy_to_vtk(da.values.reshape(1,-1).T, deep=True, array_type=VTK_FLOAT)
                    array.SetName(da.name)
                    sgrid.GetPointData().AddArray(array)
                else:
                    print (f'ERROR: Unsupported bands count (should be 1,3 or 4) {bands} for variable {data_var}')
                    return
            else:
                array = vn.numpy_to_vtk(da.values.ravel(), deep=True, array_type=VTK_FLOAT)
                array.SetName(da.name)
                sgrid.GetPointData().AddArray(array)

        for coord in dataset.coords:
            #print (coord, dataset[coord].dims, len(dataset[coord].dims))
            if len(dataset[coord].dims) > 0:
                # real coordinate, ignore it
                continue
            #print (coord, dataset[coord].dtype)
            #print ('np.datetime64', np.issubdtype(dataset[coord].dtype, np.datetime64))
            #print ('np.int64', np.issubdtype(dataset[coord].dtype, np.int64))
            #print ('np.float64', np.issubdtype(dataset[coord].dtype, np.float64))
            if np.issubdtype(dataset[coord].dtype, np.datetime64):
                data_array = vtkStringArray()
                data_value = str(dataset[coord].dt.date.values)
            elif np.issubdtype(dataset[coord].dtype, str):
                data_array = vtkStringArray()
                data_value = str(dataset[coord].values)
            elif np.issubdtype(dataset[coord].dtype, np.int64):
                data_array = vtkIntArray()
                data_value = dataset[coord].values.astype(np.int64)
            elif np.issubdtype(dataset[coord].dtype, np.float64):
                data_array = vtkFloatArray
                data_value = dataset[coord].values.astype(np.float64)
            else:
                print(f'NOTE: unsupported attribute {coord} datatype {dataset[coord].dtype}, miss it')
                continue
            data_array.SetName(coord)
            data_array.InsertNextValue(data_value)
            sgrid.GetFieldData().AddArray(data_array)

        # required for 3D interactive rendering on Google Colab
        return sgrid

    def export_geotiff(self, data, name, caption='Exporting WGS84 GeoTIFF(s)'):
        """
        Export single GeoTIFF file "velocity.tif":
        sbas.export_geotiff(velocity, 'velocity')
        
        Export pair-based stack of GeoTIFF files like corr.2024-01-01_2024-01-02.tif, ...:
        sbas.export_geotiff(corr, 'corr')
        
        Export date-based stack of GeoTIFF files (ike disp.2024-01-01.tif, ...:
        sbas.export_geotiff(disp, 'disp')

        The exported GeoTIFF files can be converted to KMZ for Google Earth Engine using GDAL tools:
        gdalwarp -of KMLSUPEROVERLAY -co FORMAT=PNG velocity.tif velocity.kmz
        """
        import xarray as xr
        import numpy as np
        from tqdm.auto import tqdm

        assert isinstance(data, xr.DataArray), 'Argument data is not an xr.DataArray object'

        # determine if data has a stack dimension and what it is
        stackvar = data.dims[0] if len(data.dims) == 3 else None

        # prepare the progress bar
        with tqdm(desc=caption, total=len(data[stackvar]) if stackvar is not None else 1) as pbar:
            for grid in data.transpose(stackvar, ...) if stackvar is not None else [data]:
                # convert the stack variable value to a string suitable for filenames
                if stackvar is not None and np.issubdtype(grid[stackvar].dtype, np.datetime64):
                    stackval = grid[stackvar].dt.date.item()
                elif stackvar is not None:
                    stackval = grid[stackvar].astype(str).item().replace(' ', '_')
                else:
                    stackval = ''
                #print ('stackval', stackval)
                filename = f'{name}.{stackval}.tif' if stackvar is not None else f'{name}.tif'
                #print ('filename', filename)
                # convert the data to geographic coordinates if necessary and export to GeoTIFF
                self.as_geo(self.ra2ll(grid) if not self.is_geo(grid) else grid).rio.to_raster(filename)
                pbar.update(1)

    def export_geojson(self, data, name, caption='Exporting WGS84 GeoJSON', pivotal=True, digits=1):
        """
        Export GeoJSON file "velocity.geojson":
        sbas.export_geojson(velocity, 'velocity')
        """
        import xarray as xr
        import geopandas as gpd
        import numpy as np
        from tqdm.auto import tqdm
        # disable "distributed.utils_perf - WARNING - full garbage collections ..."
        from dask.distributed import utils_perf
        utils_perf.disable_gc_diagnosis()

        assert isinstance(data, xr.DataArray), 'Argument data is not an xr.DataArray object'

        # determine if data has a stack dimension and what it is
        stackvar = data.dims[0] if len(data.dims) == 3 else None

        # convert the data to geographic coordinates if necessary
        if not self.is_geo(data):
            grid = self.ra2ll(data)
        else:
            grid = data

        # split to equal chunks and rest
        lats_blocks = np.array_split(np.arange(grid.lat.size), np.arange(0, grid.lat.size, self.netcdf_chunksize//2)[1:])
        lons_blocks = np.array_split(np.arange(grid.lon.size), np.arange(0, grid.lon.size, self.netcdf_chunksize//2)[1:])

        empty = True
        # prepare the progress bar
        with tqdm(desc=caption, total=len(lats_blocks)*len(lons_blocks)) as pbar:
            with open(f'{name}.geojson', 'w') as f:
                # GeoJSON header
                f.write('{"type": "FeatureCollection", "features": [\n')
                for lats_block in lats_blocks:
                    for lons_block in lons_blocks:
                        block = grid.isel(lat=lats_block, lon=lons_block).compute()
                        df = block.to_dataframe().dropna().reset_index()
                        del block
                        df[data.name] = df[data.name].apply(lambda x: float(f"{x:.{digits}f}"))
                        for col in df.columns:
                            if np.issubdtype(df[col].dtype, np.datetime64):
                                df[col] = df[col].dt.date.astype(str)
                        if stackvar is not None and np.issubdtype(df[stackvar].dtype, np.datetime64):
                            df[stackvar] = df[stackvar].dt.date.astype(str)
                        if stackvar is not None and pivotal:
                            df = df.pivot_table(index=['lat', 'lon'], columns=stackvar,
                                                values=data.name, fill_value=np.nan).reset_index()
                        # convert to geodataframe with value and geometry columns
                        gdf = gpd.GeoDataFrame(df, geometry=gpd.points_from_xy(df.lon.round(6), df.lat.round(6)))
                        del df, gdf['lat'], gdf['lon']
                        if len(gdf):
                            # crop GeoJSON header and footer and split lines
                            chunk_str = gdf.to_json(drop_id=True)[43:-2].replace('}}, ', '}},\n')
                            f.write(chunk_str if empty else ',' + chunk_str)
                            empty = False
                            del chunk_str
                        del gdf
                        pbar.update(1)
                # GeoJSON footer
                f.write(']}')
        del grid

    def export_csv(self, data, name, caption='Exporting WGS84 CSV', digits=1, delimiter=','):
        """
        Export CSV file "velocity.csv":
        sbas.export_csv(velocity, 'velocity')
        """
        import xarray as xr
        import numpy as np
        from tqdm.auto import tqdm
        # disable "distributed.utils_perf - WARNING - full garbage collections ..."
        from dask.distributed import utils_perf
        utils_perf.disable_gc_diagnosis()
    
        assert isinstance(data, xr.DataArray), 'Argument data is not an xr.DataArray object'
    
        # determine if data has a stack dimension and what it is
        stackvar = data.dims[0] if len(data.dims) == 3 else None
    
        # convert the data to geographic coordinates if necessary
        if not self.is_geo(data):
            grid = self.ra2ll(data)
        else:
            grid = data
    
        # split to equal chunks and rest
        lats_blocks = np.array_split(np.arange(grid.lat.size), np.arange(0, grid.lat.size, self.netcdf_chunksize//2)[1:])
        lons_blocks = np.array_split(np.arange(grid.lon.size), np.arange(0, grid.lon.size, self.netcdf_chunksize//2)[1:])
    
        # prepare the progress bar
        with tqdm(desc=caption, total=len(lats_blocks)*len(lons_blocks)) as pbar:
            with open(f'{name}.csv', 'w') as f:
                # CSV header
                f.write(delimiter.join(filter(None, [stackvar, 'lon', 'lat', data.name])) + '\n')
                for lats_block in lats_blocks:
                    for lons_block in lons_blocks:
                        block = grid.isel(lat=lats_block, lon=lons_block).compute()
                        block_val = block.round(digits).values
                        block_lat = block.lat.round(6).values
                        block_lon = block.lon.round(6).values
                        if stackvar is not None:
                            stackvals = block[stackvar]
                            if np.issubdtype(stackvals.dtype, np.datetime64):
                                stackvals = stackvals.dt.date.astype(str)                           
                            stackvals, lats, lons = np.meshgrid(stackvals, block_lat, block_lon, indexing='ij')
                            block_csv = np.column_stack((stackvals.ravel(), lons.ravel(), lats.ravel(), block_val.ravel()))
                            del stackvals, lats, lons
                        else:
                            lats, lons = np.meshgrid(block_lat, block_lon, indexing='ij')
                            block_csv = np.column_stack((lons.ravel(), lats.ravel(), block_val.astype(str).ravel()))
                            del lats, lons
                        del block, block_lat, block_lon
                        block_csv = block_csv[np.isfinite(block_val.ravel())]
                        del block_val
                        if block_csv.size > 0:
                            np.savetxt(f, block_csv, delimiter=delimiter, fmt='%s')
                        del block_csv
                        pbar.update(1)
        del grid

    def export_netcdf(self, data, name, caption='Exporting WGS84 NetCDF', engine='netcdf4', format='NETCDF3_64BIT'):
        """
        Export NetCDF file "velocity.nc":
        sbas.export_netcdf(velocity, 'velocity')
        """
        import xarray as xr
        import numpy as np
        import dask
        import os
        # disable "distributed.utils_perf - WARNING - full garbage collections ..."
        from dask.distributed import utils_perf
        utils_perf.disable_gc_diagnosis()
    
        assert isinstance(data, xr.DataArray), 'Argument data is not an xr.DataArray object'
    
        # convert the data to geographic coordinates if necessary
        if not self.is_geo(data):
            grid = self.ra2ll(data)
        else:
            grid = data
    
        filename = f'{name}.nc'
        if os.path.exists(filename):
            os.remove(filename)
        encoding = {data.name: self._compression(grid.shape)}
        delayed = grid.to_netcdf(filename, engine=engine, encoding=encoding, format=format, compute=False)
        tqdm_dask(dask.persist(delayed), desc=caption)
        del grid

    def export_vtk(self, data, name, caption='Exporting WGS84 VTK(s)', topo='auto', image=None, mask=None):
        """
        Export VTK file "velocity.vtk":
        sbas.export_vtk(velocity, 'velocity')
        """
        import xarray as xr
        import numpy as np
        from vtk import vtkStructuredGridWriter, vtkStringArray, VTK_BINARY
        from tqdm.auto import tqdm
        import os
    
        assert data is None or isinstance(data, xr.DataArray), 'Argument data is not an xr.DataArray object or None'
        assert data is not None or image is not None, 'One of arguments "data" or "image" needs to be specified'
        
        # determine if data has a stack dimension and what it is
        stackvar = data.dims[0] if data is not None and len(data.dims) == 3 else None
        #print ('stackvar', stackvar)
        if stackvar is not None and np.issubdtype(data[stackvar].dtype, np.datetime64):
            stackvals = data[stackvar].dt.date.astype(str).values
        elif stackvar is not None:
            stackvals = data[stackvar].astype(str).values
        else:
            stackvals = [None]
        #print ('stackvals', stackvals)
    
        filename = f'{name}.vtk'
        if os.path.exists(filename):
            os.remove(filename)
    
        dss = []
        grid2d = None
    
        if data is not None:
            # convert the data to geographic coordinates if necessary
            data_ll = self.ra2ll(data) if not self.is_geo(data) else data
            # define 2D grid for interpolation 
            grid2d = data_ll.min(stackvar) if stackvar is not None else data_ll
            dss.append(data_ll)
        
        if image is not None:
            image_ll = self.ra2ll(image) if not self.is_geo(image) else image
            if not 'band' in image.dims:
                image_ll = image_ll.expand_dims('band')
            if grid2d is not None:
                image_ll = image_ll.interp_like(grid2d, method='linear')
                #.round().astype(np.uint8)
            else:
                grid2d = image_ll.isel(band=0)
            dss.append(image_ll.rename('colors'))
        
        #print ('dss', dss)
        #print ('grid2d', grid2d)
        
        if isinstance(topo, str) and topo == 'auto':
            dem = self.get_dem()
        elif topo is not None:
            # convert topography to geographic coordinates if necessary
            dem = self.ra2ll(topo) if not self.is_geo(topo) else topo
        if topo is not None:
            dem = dem.interp_like(grid2d, method='linear')
            if isinstance(mask, str) and mask == 'auto':
                dem = dem.where(np.isfinite(grid2d))
            elif mask is not None:
                dem = dem.where(mask.reindex_like(grid2d, method='nearest'))
            dss.append(dem.rename('z'))
    
        ds = xr.merge(dss, compat='override').rename({'lat': 'y', 'lon': 'x'})
        #print ('ds', ds)
    
        # prepare the progress bar
        with tqdm(desc=caption, total=len(stackvals)) as pbar:
            for stackidx, stackval in enumerate(stackvals):
                #print (stackidx, stackval)
                if stackval is not None:
                    out = ds.isel({stackvar: stackidx}).drop_vars(stackvar)
                    filename = f'{name}.{stackidx}.vtk'
                else:
                    out = ds
                    filename = f'{name}.vtk'
                #print ('out', out)
    
                vtk_grid = self.as_vtk(out)
                if stackval is not None:
                    metadata = vtkStringArray()
                    metadata.SetName(stackvar)
                    metadata.InsertNextValue(stackval)
                    vtk_grid.GetFieldData().AddArray(metadata)
    
                # convert to VTK structure and save to file
                writer = vtkStructuredGridWriter()
                writer.SetFileName(filename)
                writer.SetInputData(vtk_grid)
                writer.SetFileType(VTK_BINARY)
                writer.Write()
                pbar.update(1)



# ----------------------------------------------------------------------------
# PyGMTSAR
# 
# This file is part of the PyGMTSAR project: https://github.com/mobigroup/gmtsar
# 
# Copyright (c) 2021, Alexey Pechnikov
# 
# Licensed under the BSD 3-Clause License (see LICENSE for details)
# ----------------------------------------------------------------------------
from .Stack_prm import Stack_prm

class Stack_orbits(Stack_prm):
    # for precision orbit there is only single orbit per day
    # for approximate orbit 2 and maybe more orbits per day are possible
    def download_orbits(self, strict=False):
        """
        Download missed orbits for all the Stack scenes.

        Returns
        -------
        None

        Examples
        --------
        stack.download_orbits()
        """
        #from eof.download import download_eofs

        print ('NOTICE: this function is removed due to continuous downloading issues from free access servers.')
        print ('NOTICE: use ASF module to download scenes and orbits.')
        print ('NOTICE: or call "eof" command-line tool in your data directory.')

#         if not strict:
#             # not strict (fast) mode — download a single orbit file for (mission, date) pairs
#             # for all (date, mission) unique pairs download the corresponding orbit files
#             df = self.df[self.df['orbitpath'].isna()][['datetime','mission']].reset_index()\
#                 .drop_duplicates(['date','mission']).sort_values('datetime')
#             if len(df) == 0:
#                 return
#             orbitpaths = download_eofs(df.datetime.tolist(), df.mission.tolist(), save_dir=self.basedir)
#             # dataframe is already sorted
#             df['orbitpath'] = sorted(orbitpaths, key=lambda x: x[-19:])
#             for record in df.itertuples():
#                 self.df.loc[(self.df.index == record.date)&(self.df.mission == record.mission),'orbitpath'] = record.orbitpath
#         else:
#             # strict mode — iterate all the records and download all the orbits separately
#             df = self.df[self.df['orbitpath'].isna()]
#             # download all the misssed orbit files
#             for record in df.itertuples():
#                 #print (date, mission)
#                 orbitpath = download_eofs([record.datetime], [record.mission], save_dir=self.basedir)[0]
#                 #print ('orbitpath', orbitpath)
#                 self.df.loc[self.df.datetime == record.datetime,'orbitpath'] = orbitpath



# ----------------------------------------------------------------------------
# PyGMTSAR
# 
# This file is part of the PyGMTSAR project: https://github.com/mobigroup/gmtsar
# 
# Copyright (c) 2023, Alexey Pechnikov
# 
# Licensed under the BSD 3-Clause License (see LICENSE for details)
# ----------------------------------------------------------------------------
__version__ = '2024.4.17.post8'

# unified progress indicators
from .tqdm_joblib import tqdm_joblib
from .tqdm_dask import tqdm_dask
# base NetCDF operations and parameters on NetCDF grid
from .datagrid import datagrid
# top level module classes
from .PRM import PRM
from .Stack import Stack
# Sentinel-1 processing functions
from .S1 import S1
# export to VTK format
from .NCubeVTK import NCubeVTK
# ASF, AWS, ESA, GMT downloading functions
from .ASF import ASF
from .AWS import AWS
from .GMT import GMT
# tiles downloading
from .Tiles import Tiles
# XYZ map tiles downloading
from .XYZTiles import XYZTiles
# morphology and other helper functions
from .utils import utils



# ----------------------------------------------------------------------------
# PyGMTSAR
# 
# This file is part of the PyGMTSAR project: https://github.com/mobigroup/gmtsar
# 
# Copyright (c) 2023, Alexey Pechnikov
# 
# Licensed under the BSD 3-Clause License (see LICENSE for details)
# ----------------------------------------------------------------------------
from .Stack_trans_inv import Stack_trans_inv

class Stack_topo(Stack_trans_inv):

    def get_topo(self):
        """
        Get the radar topography grid.

        Returns
        -------
        xarray.DataArray
            The 'topo' variable data is a xarray.DataArray.

        Examples
        --------
        Get DEM:
        topo = stack.get_topo()

        Notes
        -----
        This method returns 'topo' variable from inverse radar transform grid.
        """
        return self.get_trans_inv()['ele'].rename('topo')

    def plot_topo(self, topo='auto', caption='Topography on WGS84 ellipsoid, [m]',
                  quantile=None, vmin=None, vmax=None, symmetrical=False,
                  cmap='gray', aspect=None, **kwargs):
        import numpy as np
        import matplotlib.pyplot as plt

        if isinstance(topo, str) and topo == 'auto':
            topo = self.get_topo()
            
        if quantile is not None:
            assert vmin is None and vmax is None, "ERROR: arguments 'quantile' and 'vmin', 'vmax' cannot be used together"

        if quantile is not None:
            vmin, vmax = np.nanquantile(topo, quantile)

        # define symmetrical boundaries
        if symmetrical is True and vmax > 0:
            minmax = max(abs(vmin), vmax)
            vmin = -minmax
            vmax =  minmax

        plt.figure()
        topo.plot.imshow(cmap=cmap, vmin=vmin, vmax=vmax)
        self.plot_AOI(**kwargs)
        self.plot_POI(**kwargs)
        if aspect is not None:
            plt.gca().set_aspect(aspect)
        plt.xlabel('Range')
        plt.ylabel('Azimuth')
        plt.title(caption)

    def topo_phase(self, pairs, topo='auto', debug=False):
        """
        decimator = sbas.decimator(resolution=15, grid=(1,1))
        topophase = sbas.topo_phase(pairs, topo=decimator(sbas.get_topo()))
        """
        import pandas as pd
        import dask
        import dask.dataframe
        import xarray as xr
        import numpy as np
        #from tqdm.auto import tqdm
        import joblib
        import warnings
        # suppress Dask warning "RuntimeWarning: invalid value encountered in divide"
        warnings.filterwarnings('ignore')
        warnings.filterwarnings('ignore', module='dask')
        warnings.filterwarnings('ignore', module='dask.core')

        if debug:
            print ('DEBUG: topo_phase')

        # convert pairs (list, array, dataframe) to 2D numpy array
        pairs, dates = self.get_pairs(pairs, dates=True)
        pairs = pairs[['ref', 'rep']].astype(str).values

        if isinstance(topo, str) and topo == 'auto':
            topo = self.get_topo()

        # calculate the combined earth curvature and topography correction
        def calc_drho(rho, topo, earth_radius, height, b, alpha, Bx):
            sina = np.sin(alpha)
            cosa = np.cos(alpha)
            c = earth_radius + height
            # compute the look angle using equation (C26) in Appendix C
            # GMTSAR uses long double here
            #ret = earth_radius + topo.astype(np.longdouble)
            ret = earth_radius + topo
            cost = ((rho**2 + c**2 - ret**2) / (2. * rho * c))
            #if (cost >= 1.)
            #    die("calc_drho", "cost >= 0");
            sint = np.sqrt(1. - cost**2)
            # Compute the offset effect from non-parallel orbit
            term1 = rho**2 + b**2 - 2 * rho * b * (sint * cosa - cost * sina) - Bx**2
            drho = -rho + np.sqrt(term1)
            del term1, sint, cost, ret, c, cosa, sina
            return drho

        def block_phase(prm1, prm2, ylim, xlim):
            # use outer variables date, stack_prm
            # disable "distributed.utils_perf - WARNING - full garbage collections ..."
            from dask.distributed import utils_perf
            utils_perf.disable_gc_diagnosis()
            import warnings
            # suppress Dask warning "RuntimeWarning: invalid value encountered in divide"
            warnings.filterwarnings('ignore')
            warnings.filterwarnings('ignore', module='dask')
            warnings.filterwarnings('ignore', module='dask.core')

            # for lazy Dask ddataframes
            #prm1 = PRM(prm1)
            #prm2 = PRM(prm2)
            #prm1,  prm2  = stack_prm[stack_idx]
            #data1, data2 = stack_data[stack_idx]

            # use outer variables topo, data1, data2, prm1, prm2
            # build topo block
            block_topo = topo.isel(y=slice(ylim[0], ylim[1]), x=slice(xlim[0], xlim[1]))\
                        .compute(n_workers=1)\
                        .fillna(0)

            # set dimensions
            xdim = prm1.get('num_rng_bins')
            ydim = prm1.get('num_patches') * prm1.get('num_valid_az')

            # set heights
            htc = prm1.get('SC_height')
            ht0 = prm1.get('SC_height_start')
            htf = prm1.get('SC_height_end')

            # compute the time span and the time spacing
            tspan = 86400 * abs(prm2.get('SC_clock_stop') - prm2.get('SC_clock_start'))
            assert (tspan >= 0.01) and (prm2.get('PRF') >= 0.01), 'Check sc_clock_start, sc_clock_end, or PRF'

            from scipy import constants
            # setup the default parameters
            # constant from GMTSAR code for consistency
            #SOL = 299792456.0
            drange = constants.speed_of_light / (2 * prm2.get('rng_samp_rate'))
            #drange = SOL / (2 * prm2.get('rng_samp_rate'))
            alpha = prm2.get('alpha_start') * np.pi / 180
            cnst = -4 * np.pi / prm2.get('radar_wavelength')

            # calculate initial baselines
            Bh0 = prm2.get('baseline_start') * np.cos(prm2.get('alpha_start') * np.pi / 180)
            Bv0 = prm2.get('baseline_start') * np.sin(prm2.get('alpha_start') * np.pi / 180)
            Bhf = prm2.get('baseline_end')   * np.cos(prm2.get('alpha_end')   * np.pi / 180)
            Bvf = prm2.get('baseline_end')   * np.sin(prm2.get('alpha_end')   * np.pi / 180)
            Bx0 = prm2.get('B_offset_start')
            Bxf = prm2.get('B_offset_end')

            # first case is quadratic baseline model, second case is default linear model
            if prm2.get('baseline_center') != 0 or prm2.get('alpha_center') != 0 or prm2.get('B_offset_center') != 0:
                Bhc = prm2.get('baseline_center') * np.cos(prm2.get('alpha_center') * np.pi / 180)
                Bvc = prm2.get('baseline_center') * np.sin(prm2.get('alpha_center') * np.pi / 180)
                Bxc = prm2.get('B_offset_center')

                dBh = (-3 * Bh0 + 4 * Bhc - Bhf) / tspan
                dBv = (-3 * Bv0 + 4 * Bvc - Bvf) / tspan
                ddBh = (2 * Bh0 - 4 * Bhc + 2 * Bhf) / (tspan * tspan)
                ddBv = (2 * Bv0 - 4 * Bvc + 2 * Bvf) / (tspan * tspan)

                dBx = (-3 * Bx0 + 4 * Bxc - Bxf) / tspan
                ddBx = (2 * Bx0 - 4 * Bxc + 2 * Bxf) / (tspan * tspan)
            else:
                dBh = (Bhf - Bh0) / tspan
                dBv = (Bvf - Bv0) / tspan
                dBx = (Bxf - Bx0) / tspan
                ddBh = ddBv = ddBx = 0

            # calculate height increment
            dht = (-3 * ht0 + 4 * htc - htf) / tspan
            ddht = (2 * ht0 - 4 * htc + 2 * htf) / (tspan * tspan)

            # multiply xr.ones_like(topo) for correct broadcasting
            near_range = xr.ones_like(block_topo)*(prm1.get('near_range') + \
                block_topo.x * (1 + prm1.get('stretch_r')) * drange) + \
                xr.ones_like(block_topo)*(block_topo.y * prm1.get('a_stretch_r') * drange)

            # calculate the change in baseline and height along the frame if topoflag is on
            time = block_topo.y * tspan / (ydim - 1)        
            Bh = Bh0 + dBh * time + ddBh * time**2
            Bv = Bv0 + dBv * time + ddBv * time**2
            Bx = Bx0 + dBx * time + ddBx * time**2
            B = np.sqrt(Bh * Bh + Bv * Bv)
            alpha = np.arctan2(Bv, Bh)
            height = ht0 + dht * time + ddht * time**2

            # calculate the combined earth curvature and topography correction
            drho = calc_drho(near_range, block_topo, prm1.get('earth_radius'), height, B, alpha, Bx)

            # make topographic and model phase corrections
            # GMTSAR uses float32 complex operations with precision loss
            #phase_shift = np.exp(1j * (cnst * drho).astype(np.float32))
            phase_shift = np.exp(1j * (cnst * drho))
            del block_topo, near_range, drho, height, B, alpha, Bx, Bv, Bh, time

            return phase_shift.astype(np.complex64)

    #     # prepare lazy PRM
    #     # this is the "true way" but processing is ~40% slower due to additional Dask tasks
    #     def block_prms(date1, date2):
    #         prm1 = self.PRM(date1)
    #         prm2 = self.PRM(date2)
    #         prm2.set(prm1.SAT_baseline(prm2, tail=9)).fix_aligned()
    #         prm1.set(prm1.SAT_baseline(prm1).sel('SC_height','SC_height_start','SC_height_end')).fix_aligned()
    #         return (prm1.df, prm2.df)
    #     # Define metadata explicitly to match PRM dataframe
    #     prm_meta = pd.DataFrame(columns=['name', 'value']).astype({'name': 'str', 'value': 'object'}).set_index('name')

        # immediately prepare PRM
        # here is some delay on the function call but the actual processing is faster
        def prepare_prms(pair):
            date1, date2 = pair
            prm1 = self.PRM(date1)
            prm2 = self.PRM(date2)
            prm2.set(prm1.SAT_baseline(prm2, tail=9)).fix_aligned()
            prm1.set(prm1.SAT_baseline(prm1).sel('SC_height','SC_height_start','SC_height_end')).fix_aligned()
            return {(date1, date2): (prm1, prm2)}

        #with self.tqdm_joblib(tqdm(desc=f'Pre-Processing PRM', total=len(pairs))) as progress_bar:
        prms = joblib.Parallel(n_jobs=-1)(joblib.delayed(prepare_prms)(pair) for pair in pairs)
        # convert the list of dicts to a single dict
        prms = {k: v for d in prms for k, v in d.items()}

        # define blocks
        chunks = topo.chunks
        ychunks, xchunks = chunks[0], chunks[1]
        ychunks = np.concatenate([[0], np.cumsum(ychunks)])
        xchunks = np.concatenate([[0], np.cumsum(xchunks)])
        ylims = [(y1, y2) for y1, y2 in zip(ychunks, ychunks[1:])]
        xlims = [(x1, x2) for x1, x2 in zip(xchunks, xchunks[1:])]
        #print ('ylims', ylims)
        #print ('xlims', xlims)

        stack = []
        for stack_idx, pair in enumerate(pairs):
            date1, date2 = pair

            # Create a Dask DataFrame with provided metadata for each Dask block
            #prms = dask.delayed(block_prms)(date1, date2)
            #prm1 = dask.dataframe.from_delayed(dask.delayed(prms[0]), meta=prm_meta)
            #prm2 = dask.dataframe.from_delayed(dask.delayed(prms[1]), meta=prm_meta)
            prm1, prm2 = prms[(date1, date2)]

            blocks_total = []
            for ylim in ylims:
                blocks = []
                for xlim in xlims:
                    delayed = dask.delayed(block_phase)(prm1, prm2, ylim, xlim)
                    block = dask.array.from_delayed(delayed,
                                                    shape=((ylim[1]-ylim[0]), (xlim[1]-xlim[0])),
                                                    dtype=np.complex64)
                    blocks.append(block)
                    del block, delayed
                blocks_total.append(blocks)
                del blocks
            intf = xr.DataArray(dask.array.block(blocks_total), coords={'y': topo.y, 'x': topo.x})
            del blocks_total

            # add to stack
            stack.append(intf)
            # cleanup
            del intf, prm1, prm2
        del prms

        coord_pair = [' '.join(pair) for pair in pairs]
        coord_ref = xr.DataArray(pd.to_datetime(pairs[:,0]), coords={'pair': coord_pair})
        coord_rep = xr.DataArray(pd.to_datetime(pairs[:,1]), coords={'pair': coord_pair})

        return xr.concat(stack, dim='pair').assign_coords(ref=coord_ref, rep=coord_rep, pair=coord_pair).where(np.isfinite(topo)).rename('phase')

    def topo_slope(self, topo='auto', edge_order=1):
        import xarray as xr
        import numpy as np

        if isinstance(topo, str) and topo == 'auto':
            topo = self.get_topo()

        dy, dx = self.get_spacing(topo)
        # The cell sizes (dx, dy) are passed to np.gradient to account for non-uniform spacing
        gradient_y, gradient_x = np.gradient(topo, dy, dx, edge_order=edge_order)
        slope_radians = np.arctan(np.sqrt(gradient_x**2 + gradient_y**2))
        slope_degrees = np.degrees(slope_radians)
        slope = xr.DataArray(slope_degrees, coords=topo.coords, dims=topo.dims, name="slope_degrees")
        slope.attrs["units"] = "degrees"
        slope.attrs["description"] = "Slope calculated from DEM, accounting for non-uniform grid spacing"

        return slope

    def topo_variation(self):
        """
        Topography related range cell variation to detect layovers and shadows.
        """
        drange = self.get_trans_inv().ll.diff('x')
        return drange/drange.mean()
